from .invert import *

from operator import mul
from functools import reduce
from dolo.numeric.optimize.newton import SerialDifferentiableFunction


def prod(l):
    return reduce(mul, l)


from math import sqrt
from numba import jit
import time
from numpy import array, zeros
from numpy.linalg import inv, solve
from numpy.linalg import inv
from numpy import abs
from numba import jit

import time


def B_prod(bases):

    # Compute B for a product of bases

    import scipy
    import scipy.sparse

    n = [b.B.shape[0] - 2 for b in bases]
    dims = array(n) + 2

    if len(dims) == 1:
        B = bases[0].B

    elif len(dims) == 2:

        B_x = bases[0].B
        B_y = bases[1].B
        B_prod_x = numpy.zeros(tuple(dims) * 2)  # filters last dim
        B_prod_y = numpy.zeros(tuple(dims) * 2)  # filters first dim

        sh = B_prod_x.shape

        for i in range(B_prod_x.shape[1]):
            # for j in B_prod_x.shape[3]:
            B_prod_x[:, i, :, i] = B_x

        for i in range(B_prod_y.shape[0]):
            # for j in B_prod_x.shape[3]
            B_prod_y[i, :, i, :] = B_y

        dim = numpy.prod(dims)

        B_p_x = scipy.sparse.csc_matrix(B_prod_x.reshape((dim, dim)))
        B_p_y = scipy.sparse.csc_matrix(B_prod_y.reshape((dim, dim)))

        B = B_p_x @ B_p_y

        B = scipy.sparse.csc_matrix(B)

    elif len(dims) == 3:

        B_x = bases[0].B
        B_y = bases[1].B
        B_z = bases[2].B
        B_prod_x = numpy.zeros(tuple(dims) * 2)  # filters last dim
        B_prod_y = numpy.zeros(tuple(dims) * 2)  # filters first dim
        B_prod_z = numpy.zeros(tuple(dims) * 2)  # filters first dim

        sh = B_prod_x.shape

        for i in range(B_prod_x.shape[1]):
            for j in range(B_prod_x.shape[2]):
                # for j in B_prod_x.shape[3]:
                B_prod_x[:, i, j, :, i, j] = B_x

        for i in range(B_prod_y.shape[0]):
            for j in range(B_prod_x.shape[2]):
                # for j in B_prod_x.shape[3]:
                B_prod_y[i, :, j, i, :, j] = B_y

        for i in range(B_prod_y.shape[0]):
            for j in range(B_prod_x.shape[1]):
                # for j in B_prod_x.shape[3]:
                B_prod_z[i, j, :, i, j, :] = B_z

        dim = numpy.prod(dims)

        B_p_x = scipy.sparse.csc_matrix(B_prod_x.reshape((dim, dim)))
        B_p_y = scipy.sparse.csc_matrix(B_prod_y.reshape((dim, dim)))
        B_p_z = scipy.sparse.csc_matrix(B_prod_z.reshape((dim, dim)))

        # C_p = (B_p_y@B_p_x)
        # C_p = (B_p_x@B_p_y)
        B = B_p_x @ B_p_y @ B_p_z
        # C_p = (B_p_z@B_p_y@B_p_x)

    return B


@jit(nopython=True)
def mult_AB(dres, jres):
    n_m, N, n_x, n_xx = dres.shape
    assert jres.shape == (n_m, N, n_x, n_m, n_x)
    out = jres.copy()
    for i in range(n_m):
        for j in range(n_m):
            for v in range(n_x):
                out[i, :, :, j, v] = serial_solve(dres[i, :, :, :], jres[i, :, :, j, v])
    return out


class SparseTensor:
    def __init__(self, indices, values, shape):
        """
        indices: N x d integers
        values:  N     floats
        """
        indices = numpy.array(indices, dtype=int)
        values = numpy.array(values, dtype=float)
        self.indices = indices
        self.values = values
        self.shape = shape

    @property
    def lin_indices(self):
        minds = tuple(self.indices.T)
        return numpy.ravel_multi_index(minds, self.shape)

    def reshape(self, shape):
        assert prod(shape) == prod(self.shape)
        new_indices = numpy.unravel_index(self.lin_indices, shape)
        new_indices = numpy.concatenate([a[:, None] for a in new_indices], axis=1)
        return SparseTensor(new_indices, self.values, shape)

    def as_spmatrix(self, dims=None):
        import scipy.sparse

        if dims is None:
            pp = prod(self.shape)
            P = int(sqrt(pp))
            Q = P
            assert pp == P ** 2
            dims = (P, Q)
        else:
            assert len(dims) == 2
        spmat = self.reshape(dims)
        data = spmat.values
        inds = spmat.indices
        return scipy.sparse.coo_matrix((data, (inds[:, 0], inds[:, 1])))

    def as_array(self, dims=None):
        return self.as_spmatrix(dims).toarray()


def dres_to_sparse(dres):
    import numpy

    n_m, N, n_x, junk = dres.shape
    assert n_x == junk
    big_N = n_m * N * n_x * n_m * N * n_x
    nnz = n_m * N * n_x * n_x
    inds = numpy.zeros((nnz, 6), dtype=int)
    vals = numpy.zeros(nnz, dtype=float)
    n = 0
    for i in range(n_m):
        for j in range(N):
            for k in range(n_x):
                for l in range(n_x):
                    inds[n, 0] = i
                    inds[n, 1] = j
                    inds[n, 2] = k
                    inds[n, 3] = i
                    inds[n, 4] = j
                    inds[n, 5] = l
                    vals[n] = dres[i, j, k, l]
                    n += 1
    return SparseTensor(inds, vals, (n_m, N, n_x, n_m, N, n_x))


@jit
def serial_solve(A, B, diagnose=True):

    sol = zeros_like(B)
    for i in range(sol.shape[0]):
        sol[i, :] = solve(A[i, :, :], B[i, :])

    return sol


def jres_to_sparse(jres):
    import numpy

    n_m, N, n_x, a, b = jres.shape
    assert a == n_m
    assert b == n_x
    big_N = n_m * N * n_x * n_m * N * n_x
    nnz = n_m * N * n_x * n_m * n_x
    inds = numpy.zeros((nnz, 6), dtype=float)
    vals = numpy.zeros(nnz, dtype=float)
    n = 0
    for i in range(n_m):
        for j in range(N):
            for k in range(n_x):
                for l in range(n_m):
                    for m in range(n_x):
                        inds[n, 0] = i
                        inds[n, 1] = j
                        inds[n, 2] = k
                        inds[n, 3] = l
                        inds[n, 4] = j
                        inds[n, 5] = m
                        vals[n] = jres[i, j, k, l, m]
                        n += 1
    return SparseTensor(inds, vals, (n_m, N, n_x, n_m, N, n_x))


def compact_ddx(res, fut_S):

    import scipy.sparse
    import scipy.sparse

    n_ms, N, n_x = res.shape

    Binv = numpy.linalg.inv(B)
    #         DDX = numpy.zeros((n_ms,N,n_x,n_ms,N,n_x))
    nnz = N * N * n_m * n_m * n_x
    indices = numpy.zeros((nnz, 6), dtype=int)
    vals = numpy.zeros(nnz, dtype=float)
    n = 0
    for i_m in range(n_ms):
        for i_M in range(n_ms):
            S = fut_S[i_m, :, i_M, :]
            Phi = tb.Phi(S).as_matrix()
            DD = (Phi @ Binv)[:, 1:-1]
            #             DD[abs(DD)<1e-6] = 0
            for i_x in range(n_x):
                # Phi(S) c = Phi(S) (B^{-1}) x
                # DDX[i_m,:,i_x,i_m,:,i_x] = DD
                for p in range(N):
                    for q in range(N):
                        indices[n, 0] = i_m
                        indices[n, 1] = p
                        indices[n, 2] = i_x
                        indices[n, 3] = i_M
                        indices[n, 4] = q
                        indices[n, 5] = i_x
                        vals[n] = DD[p, q]
                        n += 1
    indices = indices[:n, :]
    vals = vals[:n]
    return SparseTensor(indices, vals, (n_ms, N, n_x, n_ms, N, n_x))


#         return scipy.sparse.coo_matrix( DDX.reshape((bigN,bigN)) )


def construct_j2(jres):
    import numpy

    m_m, N, n_x, a, b = jres.shape
    assert a == n_m
    assert b == n_x
    big_N = n_m * N * n_x * n_m * N * n_x
    nnz = m_m * N * n_x * n_m * n_x
    inds = numpy.zeros((nnz, 6), dtype=float)
    vals = numpy.zeros(nnz, dtype=float)
    n = 0
    for i in range(n_m):
        for j in range(N):
            for k in range(n_x):
                for l in range(n_m):
                    for m in range(n_x):
                        inds[n, 0] = i
                        inds[n, 1] = j
                        inds[n, 2] = k
                        inds[n, 3] = l
                        inds[n, 4] = j
                        inds[n, 5] = m
                        vals[n] = jres[i, j, k, l, m]
                        n += 1
    return SparseTensor(inds, vals, (n_m, N, n_x, n_m, N, n_x))


from numpy import sqrt


from dolo.numeric.serial_operations import serial_multiplication as smul

#
# from interpolation.linear_bases.basis_uniform_cubic_splines import UniformSplineBasis
# from interpolation.linear_base    s.product import TensorBase
from dolo.numeric.serial_operations import serial_multiplication as smult


from numpy.linalg import solve
from numpy import zeros_like


class SmartJacobian:
    def __init__(self, res, dres, jres, fut_S, grid):
        self.res = res
        self.dres = dres
        self.jres = jres
        self.fut_S = fut_S
        self.n_m = res.shape[0]
        self.N = res.shape[1]
        self.n_x = res.shape[2]
        self.grid = grid
        self.__B__ = None

    def B(self):
        if self.__B__ is None:
            self.__B__ = self.get_filter()
        return self.__B__

    def get_filter(self):
        a = self.grid.a
        b = self.grid.b
        o = self.grid.orders

        bases = [UniformSplineBasis(a[i], b[i], o[i]) for i in range(len(o))]
        tb = TensorBase(bases)
        self.tb = tb

        B = B_prod(bases)

        return B

    @property
    def jac_1(self):

        n_ms, N, n_x = self.res.shape
        dres = self.dres
        bigN = n_ms * N * n_x
        DDRes = numpy.zeros((n_ms, N, n_x, n_ms, N, n_x))
        for n in range(N):
            for i_m in range(n_ms):
                DDRes[i_m, n, :, i_m, n, :] = dres[i_m, n, :, :]
        jac1 = DDRes.reshape((bigN, bigN))
        return jac1

    @property
    def j2_A(self):

        n_ms, N, n_x = self.res.shape
        jres = self.jres
        fut_S = self.fut_S
        bigN = n_ms * N * n_x

        JJRes = numpy.zeros((n_ms, N, n_x, n_ms, N, n_x))
        for n in range(N):
            # for i_m in range(n_ms):
            JJRes[:, n, :, :, n, :] = jres[:, n, :, :, :]

        return JJRes.reshape((bigN, bigN))

    @property
    def j2_B(self):

        n_ms, N, n_x = self.res.shape
        fut_S = self.fut_S
        bigN = n_ms * N * n_x
        fut_S = self.fut_S
        DDX = numpy.zeros((n_ms, N, n_x, n_ms, N, n_x))
        for i_x in range(n_x):
            for i_m in range(n_ms):
                for i_M in range(n_ms):
                    S = fut_S[i_m, :, i_M, :]
                    Phi = tb.Phi(S).as_matrix()
                    # Phi(S) c = Phi(S) (B^{-1}) x
                    # i_m -> i_M
                    DDX[i_m, :, i_x, i_m, :, i_x] = (Phi @ numpy.linalg.inv(B))[:, 1:-1]
        #                     DDX[i_M,:,i_x,i_m,:,i_x] = (Phi @ numpy.linalg.inv(B))[:,1:-1]

        return DDX.reshape((bigN, bigN))

    #     @property
    #     def jac_2(self):
    #         return self.j2_A @ self.j2_B

    @property
    def jac_2(self):

        B = self.get_filter()
        d = len(self.tb.bases)
        mdims = tuple([b.m for b in self.tb.bases])

        # works
        n_ms, N, n_x = self.res.shape
        #         jres = self.jres
        #         fut_S = self.fut_S
        bigN = n_ms * N * n_x
        jres = self.jres

        fut_S = self.fut_S
        try:
            B = B.todense()
        except:
            pass
        Binv = numpy.linalg.inv(B)
        mat = numpy.zeros((n_ms, N, n_x, n_ms, N, n_x))

        for i in range(n_ms):
            for j in range(n_ms):
                S = fut_S[i, :, j, :]
                Phi = self.tb.Phi(S).as_matrix()
                X = Phi @ Binv  # [:,1:-1]
                X = numpy.array(X)
                X = X.reshape((X.shape[0],) + mdims)
                if d == 1:
                    X = X[:, 1:-1]
                elif d == 2:
                    X = X[:, 1:-1, 1:-1]
                elif d == 3:
                    X = X[:, 1:-1, 1:-1, 1:-1]
                X = X.reshape((X.shape[0], -1))
                XX = numpy.zeros((N, n_x, N, n_x))
                for y in range(n_x):
                    XX[:, y, :, y] = X
                XX = XX.reshape((N * n_x, N * n_x))
                ff = jres[i, :, :, j, :]
                ff_m = numpy.zeros((N, n_x, N, n_x))
                for n in range(N):
                    ff_m[n, :, n, :] = ff[n, :, :]
                m = ff_m.reshape((N * n_x, N * n_x)) @ XX
                m = m.reshape((N, n_x, N, n_x))
                mat[i, :, :, j, :, :] = m
        mat = mat.reshape((bigN, bigN))
        mat[abs(mat) < 1e-6] = 0
        import scipy.sparse

        mat = scipy.sparse.coo_matrix(mat)
        return mat

    #     @property
    #     def jac_2(self):
    #         return self.j2
    # #         return self.j2_A @ self.j2_B

    @property
    def jac(self):
        return self.jac_1 + self.jac_2

    def solve(self, rr):
        import numpy.linalg

        return numpy.linalg.solve(self.jac, rr.ravel()).reshape(rr.shape)

    def solve_sp(self, rr):
        import scipy.sparse
        from scipy.sparse.linalg import spsolve

        jj = scipy.sparse.csr_matrix(self.jac)
        res = spsolve(jj, rr.ravel())
        return res.reshape(rr.shape)

    #         return numpy.linalg.solve(self.jac, rr.ravel()).reshape(rr.shape)

    def solve_smart(
        self, rr, tol=1e-10, maxit=1000, verbose=False, filt=None, scale=1.0
    ):
        n_m, N, n_x = self.res.shape
        fut_S = self.fut_S
        bigN = n_m * N * n_x
        dres = self.dres
        jres = self.jres
        grid = self.grid
        fut_S = self.fut_S
        sol, nn = invert_jac(
            rr * scale,
            dres,
            jres,
            fut_S,
            n_m,
            N,
            n_x,
            grid,
            tol=tol,
            maxit=maxit,
            verbose=verbose,
            filt=filt,
        )
        sol /= scale
        return sol

    def solve_ind(self, rr, tol=1e-12):
        # works only if there are no shocks
        jac1 = self.jac_1
        j2_A = self.j2_A
        j2_B = self.j2_B
        M = jac1
        N = -numpy.linalg.solve(jac1, j2_A @ j2_B)
        # I need to prove sp(N)<[0,1[

        abs(numpy.linalg.eig(N)[0]).max()
        abs(numpy.linalg.eig(N)[0]).min()

        term = numpy.linalg.solve(M, rr.ravel())
        tot = term
        for i in range(10000):
            term = N @ term
            err = abs(term).max()
            tot = tot + term
            if err < tol:
                break
        return tot.reshape(rr.shape)


#### utilities to compute perturbations
from dolo.numeric.extern.qz import qzordered
from numpy.linalg import solve
import numpy as np


def classical_perturbation(g_s, g_x, f_s, f_x, f_S, f_X):
    n_s = g_s.shape[0]  # number of controls
    n_x = g_x.shape[1]  # number of states
    n_v = n_s + n_x

    A = row_stack(
        [column_stack([eye(n_s), zeros((n_s, n_x))]), column_stack([-f_S, -f_X])]
    )

    B = row_stack([column_stack([g_s, g_x]), column_stack([f_s, f_x])])

    [S, T, Q, Z, eigval] = qzordered(A, B, 1.0 - 1e-8)

    Q = Q.real  # is it really necessary ?
    Z = Z.real

    diag_S = np.diag(S)
    diag_T = np.diag(T)

    tol_geneigvals = 1e-10

    Z11 = Z[:n_s, :n_s]
    # Z12 = Z[:n_s, n_s:]
    Z21 = Z[n_s:, :n_s]
    # Z22 = Z[n_s:, n_s:]
    # S11 = S[:n_s, :n_s]
    # T11 = T[:n_s, :n_s]

    # first order solution
    # P = (solve(S11.T, Z11.T).T @ solve(Z11.T, T11.T).T)
    C = solve(Z11.T, Z21.T).T
    return C, eigval


@jit(nopython=True)
def simple_max(T):
    mm = 0.0
    p, q = T.shape
    for i in range(p):
        for j in range(q):
            e = abs(T[i, j])
            if e > mm:
                mm = e
    return mm


@jit(nopython=True)
def newtonator(g_s, g_x, f_s, f_x, f_S, f_X):  # , tol=1e-8, inner_tol=1e-8, X=None):

    # solves linear approximation using mix of time-iteration and newton algorithm

    tol = 1e-14
    inner_tol = 1e-12
    inner_maxit = 1000

    X = numpy.zeros((n_x, n_s))
    Y = X

    Ainv = numpy.zeros((n_x, n_x))
    direction = numpy.zeros_like(X)

    maxit = 1000
    err = 1.0
    it = 0

    K1 = f_s + f_S @ g_s
    K2 = f_x + f_S @ g_x
    res = K1 + K2 @ X + f_X @ Y @ g_s + f_X @ Y @ g_x @ X

    while err > tol and it < maxit:

        it += 1

        A = K2 + f_X @ Y @ g_x
        B = f_X

        Ainv = inv(A)

        C = g_s + g_x @ X
        M = -Ainv @ B
        ddx = Ainv @ res
        direction[:, :] = ddx
        for i in range(inner_maxit):
            ddx = M @ ddx @ C
            #             inner_err = (abs(ddx).max())
            #             inner_err = (abs(ddx).max())
            inner_err = simple_max(ddx)
            direction[:, :] += ddx
            if inner_err < inner_tol:
                break

        X -= direction
        Y = X
        res = K1 + K2 @ X + f_X @ Y @ g_s + f_X @ Y @ g_x @ X
        err = abs(res).max()

    #         doesn't seem to make any change
    #          err = abs(Jac).max()
    #         for t in backsteps:
    #             Y[:,:] = X[:,:] - direction[:,:]*t
    #             res = K1 + K2@Y + f_X@Y@g_s + f_X@Y@g_x@Y
    #             nerr = abs(res).max()
    #             if nerr<err:
    #                 err = nerr
    #                 break
    #         X[:,:] = Y[:,:]

    return X


def spectral_radius(g_s, g_x, f_s, f_x, f_S, f_X, X=None):

    tol = 1e-14
    inner_tol = 1e-12
    inner_maxit = 1000

    if X is None:
        X = numpy.zeros((n_x, n_s))

    Y = X

    Ainv = numpy.zeros((n_x, n_x))
    direction = numpy.zeros_like(X)

    maxit = 1000
    err = 1.0
    it = 0

    errors = []
    #     res = K1 + K2@X + f_X@Y@g_s + f_X@Y@g_x@ X

    res = numpy.random.random((n_x, n_s))

    K1 = f_s + f_S @ g_s
    K2 = f_x + f_S @ g_x
    A = K2 + f_X @ Y @ g_x
    B = f_X

    Ainv = inv(A)

    C = g_s + g_x @ X
    M = -Ainv @ B
    ddx = Ainv @ res
    for i in range(inner_maxit):
        ddx = M @ ddx @ C
        errors.append(abs(ddx.max()))

    errors = numpy.array(errors)
    #     return errors
    rats = errors[1:] / errors[:-1]
    return rats[-1]



def test_ergo_iid():

    from dolo import yaml_import, improved_time_iteration, ergodic_distribution

    model = yaml_import("examples/models/rbc_iid.yaml")
    sol = improved_time_iteration(model)
    Π, μ = ergodic_distribution(model, sol.dr)
    assert μ.ndim == 2


def test_ergo_mc():

    from dolo import yaml_import, improved_time_iteration, ergodic_distribution

    model = yaml_import("examples/models/rbc_mc.yaml")
    sol = improved_time_iteration(model)
    Π, μ = ergodic_distribution(model, sol.dr)
    assert μ.ndim == 2



class AlgoResult:
    pass


from dataclasses import dataclass


@dataclass
class TimeIterationResult(AlgoResult):
    dr: object
    iterations: int
    complementarities: bool
    dprocess: object
    x_converged: bool
    x_tol: float
    err: float
    log: object  # TimeIterationLog
    trace: object  # {Nothing,IterationTrace}


@dataclass
class ValueIterationResult(AlgoResult):
    dr: object  #:AbstractDecisionRule
    drv: object  #:AbstractDecisionRule
    iterations: int
    dprocess: object  #:AbstractDiscretizedProcess
    x_converged: object  #:Bool
    x_tol: float
    x_err: float
    v_converged: bool
    v_tol: float
    v_err: float
    log: object  #:ValueIterationLog
    trace: object  #:Union{Nothing,IterationTrace}


@dataclass
class ImprovedTimeIterationResult(AlgoResult):
    dr: object  #:AbstractDecisionRule
    N: int
    f_x: float  #:Float64
    d_x: float  #:Float64
    x_converged: bool  #:Bool
    complementarities: bool  #:Bool
    # Time_search::
    radius: float  # :Float64
    trace_data: object
    L: object


@dataclass
class PerturbationResult(AlgoResult):
    dr: object  #:BiTaylorExpansion
    generalized_eigenvalues: object  # :Vector
    stable: bool  # biggest e.v. lam of solution is < 1
    determined: bool  # next eigenvalue is > lam + epsilon (MOD solution well defined)
    unique: bool  # next eigenvalue is > 1



### The following originates from https://github.com/coady/multimethod

# Copyright 2020 Aric Coady
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

###

import abc
import collections
import functools
import inspect
import itertools
import types
import typing
from typing import Callable, Iterable, Iterator, Mapping

__version__ = "1.3"


def groupby(func: Callable, values: Iterable) -> dict:
    """Return mapping of key function to values."""
    groups = collections.defaultdict(list)  # type: dict
    for value in values:
        groups[func(value)].append(value)
    return groups


def get_types(func: Callable) -> tuple:
    """Return evaluated type hints in order."""
    if not hasattr(func, "__annotations__"):
        return ()
    annotations = dict(typing.get_type_hints(func))
    annotations.pop("return", None)
    params = inspect.signature(func).parameters
    return tuple(annotations.pop(name, object) for name in params if annotations)


class DispatchError(TypeError):
    pass


class subtype(type):
    """A normalized generic type which checks subscripts."""

    def __new__(cls, tp, *args):
        if tp is typing.Any:
            return object
        if isinstance(tp, typing.TypeVar):
            if not tp.__constraints__:
                return object
            tp = typing.Union[tp.__constraints__]
        origin = getattr(tp, "__extra__", getattr(tp, "__origin__", tp))
        args = tuple(map(cls, getattr(tp, "__args__", None) or args))
        if set(args) <= {object} and not (origin is tuple and args):
            return origin
        bases = (origin,) if type(origin) is type else ()
        namespace = {"__origin__": origin, "__args__": args}
        return type.__new__(cls, str(tp), bases, namespace)

    def __init__(self, tp, *args):
        if isinstance(self.__origin__, abc.ABCMeta):
            self.__origin__.register(self)

    def __getstate__(self):
        return self.__origin__, self.__args__

    def __eq__(self, other):
        return (
            isinstance(other, subtype) and self.__getstate__() == other.__getstate__()
        )

    def __hash__(self):
        return hash(self.__getstate__())

    def __subclasscheck__(self, subclass):
        origin = getattr(
            subclass, "__extra__", getattr(subclass, "__origin__", subclass)
        )
        args = getattr(subclass, "__args__", ())
        if origin is typing.Union:
            return all(issubclass(cls, self) for cls in args)
        if self.__origin__ is typing.Union:
            return issubclass(subclass, self.__args__)
        return (  # check args first to avoid a recursion error in ABCMeta
            len(args) == len(self.__args__)
            and issubclass(origin, self.__origin__)
            and all(map(issubclass, args, self.__args__))
        )


class signature(tuple):
    """A tuple of types that supports partial ordering."""

    parents = None  # type: set

    def __new__(cls, types: Iterable):
        return tuple.__new__(cls, map(subtype, types))

    def __le__(self, other) -> bool:
        return len(self) <= len(other) and all(map(issubclass, other, self))

    def __lt__(self, other) -> bool:
        return self != other and self <= other

    def __sub__(self, other) -> tuple:
        """Return relative distances, assuming self >= other."""
        mros = (subclass.mro() for subclass in self)
        return tuple(
            mro.index(cls if cls in mro else object) for mro, cls in zip(mros, other)
        )


class multimethod(dict):
    """A callable directed acyclic graph of methods."""

    pending = None  # type: set

    def __new__(cls, func):
        namespace = inspect.currentframe().f_back.f_locals
        self = functools.update_wrapper(dict.__new__(cls), func)
        self.pending = set()
        self.get_type = type  # default type checker
        return namespace.get(func.__name__, self)

    def __init__(self, func: Callable):
        try:
            self[get_types(func)] = func
        except NameError:
            self.pending.add(func)

    def register(self, *args):
        """Decorator for registering a function.

        Optionally call with types to return a decorator for unannotated functions.
        """
        if len(args) == 1 and hasattr(args[0], "__annotations__"):
            return overload.register(self, *args)
        return lambda func: self.__setitem__(args, func) or func

    def __get__(self, instance, owner):
        return self if instance is None else types.MethodType(self, instance)

    def parents(self, types: tuple) -> set:
        """Find immediate parents of potential key."""
        parents = {key for key in self if isinstance(key, signature) and key < types}
        return parents - {ancestor for parent in parents for ancestor in parent.parents}

    def clean(self):
        """Empty the cache."""
        for key in list(self):
            if not isinstance(key, signature):
                super().__delitem__(key)

    def __setitem__(self, types: tuple, func: Callable):
        self.clean()
        types = signature(types)
        parents = types.parents = self.parents(types)
        for key in self:
            if types < key and (not parents or parents & key.parents):
                key.parents -= parents
                key.parents.add(types)
        if any(isinstance(cls, subtype) for cls in types):
            self.get_type = get_type  # switch to slower generic type checker
        super().__setitem__(types, func)
        self.__doc__ = self.docstring

    def __delitem__(self, types: tuple):
        self.clean()
        super().__delitem__(types)
        for key in self:
            if types in key.parents:
                key.parents = self.parents(key)
        self.__doc__ = self.docstring

    def __missing__(self, types: tuple) -> Callable:
        """Find and cache the next applicable method of given types."""
        self.evaluate()
        if types in self:
            return self[types]
        groups = groupby(signature(types).__sub__, self.parents(types))
        keys = groups[min(groups)] if groups else []
        funcs = {self[key] for key in keys}
        if len(funcs) == 1:
            return self.setdefault(types, *funcs)
        msg = f"{self.__name__}: {len(keys)} methods found"  # type: ignore
        raise DispatchError(msg, types, keys)

    def __call__(self, *args, **kwargs):
        """Resolve and dispatch to best method."""
        return self[tuple(map(self.get_type, args))](*args, **kwargs)

    def evaluate(self):
        """Evaluate any pending forward references.

        This can be called explicitly when using forward references,
        otherwise cache misses will evaluate.
        """
        while self.pending:
            func = self.pending.pop()
            self[get_types(func)] = func

    @property
    def docstring(self):
        """a descriptive docstring of all registered functions"""
        docs = []
        for func in set(self.values()):
            try:
                sig = inspect.signature(func)
            except ValueError:
                sig = ""
            doc = func.__doc__ or ""
            docs.append(f"{func.__name__}{sig}\n    {doc}")
        return "\n\n".join(docs)


class multidispatch(multimethod):
    """Provisional wrapper for future compatibility with `functools.singledispatch`."""


get_type = multimethod(type)
get_type.__doc__ = """Return a generic `subtype` which checks subscripts."""
for atomic in (Iterator, str, bytes):
    get_type[
        atomic,
    ] = type


@multimethod  # type: ignore[no-redef]
def get_type(arg: tuple):
    """Return generic type checking all values."""
    return subtype(type(arg), *map(get_type, arg))


@multimethod  # type: ignore[no-redef]
def get_type(arg: Mapping):
    """Return generic type checking first item."""
    return subtype(type(arg), *map(get_type, next(iter(arg.items()), ())))


@multimethod  # type: ignore[no-redef]
def get_type(arg: Iterable):
    """Return generic type checking first value."""
    return subtype(type(arg), *map(get_type, itertools.islice(arg, 1)))


def isa(*types) -> Callable:
    """Partially bound `isinstance`."""
    return lambda arg: isinstance(arg, types)


class overload(collections.OrderedDict):
    """Ordered functions which dispatch based on their annotated predicates."""

    __get__ = multimethod.__get__

    def __new__(cls, func):
        namespace = inspect.currentframe().f_back.f_locals
        self = functools.update_wrapper(super().__new__(cls), func)
        return namespace.get(func.__name__, self)

    def __init__(self, func: Callable):
        self[inspect.signature(func)] = func

    def __call__(self, *args, **kwargs):
        """Dispatch to first matching function."""
        for sig, func in reversed(self.items()):
            arguments = sig.bind(*args, **kwargs).arguments
            if all(
                predicate(arguments[name])
                for name, predicate in func.__annotations__.items()
            ):
                return func(*args, **kwargs)
        raise DispatchError("No matching functions found")

    def register(self, func: Callable) -> Callable:
        """Decorator for registering a function."""
        self.__init__(func)  # type: ignore
        return self if self.__name__ == func.__name__ else func  # type: ignore


class multimeta(type):
    """Convert all callables in namespace to multimethods."""

    class __prepare__(dict):
        def __init__(*args):
            pass

        def __setitem__(self, key, value):
            if callable(value):
                value = getattr(self.get(key), "register", multimethod)(value)
            super().__setitem__(key, value)



from numpy import sqrt, finfo, inf
from numpy import isinf, newaxis, diag, zeros
from numpy.linalg import norm, solve
from numpy import float64
import warnings

from dolo.numeric.optimize.newton import newton


def ncpsolve(
    f, a, b, x, tol=1e-8, maxit=100, infos=False, verbose=False, jactype="serial"
):
    def fcmp(z):

        [val, dval] = f(z)

        # revert convention
        val = -val
        dval = -dval

        [val, dval] = smooth(z, a, b, val, dval, jactype=jactype)

        return [val, dval]

    [sol, nit] = newton(fcmp, x, tol=tol, maxit=maxit, verbose=verbose, jactype=jactype)

    return [sol, nit]


def smooth(x, a, b, fx, J, jactype="serial"):

    BIG = 1e20

    dainf = a <= -BIG  #  isinf(a) |
    dbinf = b >= BIG  #  isinf(b)

    da = a - x
    db = b - x

    # TODO: ignore warnings when there are infinite values.
    sq1 = sqrt(fx ** 2 + da ** 2)
    pval = fx + sq1 + da
    pval[dainf] = fx[dainf]

    sq2 = sqrt(pval ** 2 + db ** 2)

    fxnew = pval - sq2 + db

    fxnew[dbinf] = pval[dbinf]

    dpdy = 1 + fx / sq1
    dpdy[dainf] = 1

    dpdz = 1 + da / sq1
    dpdz[dainf] = 0

    dmdy = 1 - pval / sq2
    dmdy[dbinf] = 1

    dmdz = 1 - db / sq2
    dmdz[dbinf] = 0

    ff = dmdy * dpdy
    xx = dmdy * dpdz + dmdz

    # TODO: rewrite starting here

    #    jac_is_sparse = scipy.sparse.issparse(J)

    if jactype == "sparse":
        import scipy.sparse
        from scipy.sparse import csc_matrix, csr_matrix
        from scipy.sparse import diags

        fff = diags([ff], [0])
        xxx = diags([xx], [0])
        # TODO: preserve csc or csr format
        Jnew = fff * J - xxx

        Jnew = csc_matrix(Jnew)

        return [fxnew, Jnew]

        # TODf: preserve csc or csr format

    elif jactype == "full":
        from numpy import diag

        #        fff = diag(ff)
        fff = ff[:, None]
        xxx = diag(xx)
        #        xxx = diag(xx)
        Jnew = fff * J - xxx
        return fx, J
        return [fxnew, Jnew]

    else:

        fff = ff[:, :, newaxis]

        xxx = zeros(J.shape)
        for i in range(xx.shape[1]):
            xxx[:, i, i] = xx[:, i]
        Jnew = fff * J - xxx
        return [fxnew, Jnew]



# plainformatter = get_ipython().display_formatter.formatters['text/plain']
# del plainformatter.type_printers[dict]

import yaml
import numpy as np
from typing import List

import ast
from ast import BinOp, Sub

from typing import Dict

import dolang
from dolang.grammar import str_expression
from dolang.symbolic import parse_string
from dolang.symbolic import time_shift
from dolang.symbolic import Sanitizer
from dolang.factory import FlatFunctionFactory


def get_name(e):
    return ast.parse(e).body[0].value.func.id


def reorder_preamble(pr):

    from dolang.triangular_solver import triangular_solver, get_incidence

    inc = get_incidence(pr)
    order = triangular_solver(inc)
    d = dict()
    prl = [*pr.items()]
    for o in order:
        k, v = prl[o]
        d[k] = v
    return d


def shift_spec(specs, tshift):
    ss = dict()
    if "target" in specs:
        e = specs["target"]
        ss["target"] = [e[0], e[1] + tshift, e[2]]
    ss["eqs"] = [
        ([e[0], e[1] + tshift, e[2]] if e[0] != "parameters" else e)
        for e in specs["eqs"]
    ]
    return ss


def get_factory(model, eq_type: str, tshift: int = 0):

    from dolo.compiler.model import decode_complementarity

    from dolo.compiler.recipes import recipes
    from dolang.symbolic import stringify, stringify_symbol

    equations = model.equations

    if eq_type == "auxiliary":
        eqs = ["{}".format(s) for s in model.symbols["auxiliaries"]]
        specs = {
            "eqs": [
                ["exogenous", 0, "m"],
                ["states", 0, "s"],
                ["controls", 0, "x"],
                ["parameters", 0, "p"],
            ]
        }
    else:
        eqs = equations[eq_type]
        if eq_type in ("arbitrage_lb", "arbitrage_ub"):
            specs = {
                "eqs": recipes["dtcc"]["specs"]["arbitrage"]["complementarities"][
                    "left-right"
                ]
            }
        else:
            specs = recipes["dtcc"]["specs"][eq_type]

    specs = shift_spec(specs, tshift=tshift)

    preamble_tshift = set([s[1] for s in specs["eqs"] if s[0] == "states"])
    preamble_tshift = preamble_tshift.intersection(
        set([s[1] for s in specs["eqs"] if s[0] == "controls"])
    )

    args = []
    for sg in specs["eqs"]:
        if sg[0] == "parameters":
            args.append([s for s in model.symbols["parameters"]])
        else:
            args.append([(s, sg[1]) for s in model.symbols[sg[0]]])
    args = [[stringify_symbol(e) for e in vg] for vg in args]

    arguments = dict(zip([sg[2] for sg in specs["eqs"]], args))

    # temp
    eqs = [eq.split("⟂")[0].strip() for eq in eqs]

    if "target" in specs:
        sg = specs["target"]
        targets = [(s, sg[1]) for s in model.symbols[sg[0]]]
        eqs = [eq.split("=")[1] for eq in eqs]
    else:
        eqs = [
            ("({1})-({0})".format(*eq.split("=")) if "=" in eq else eq) for eq in eqs
        ]
        targets = [("out{}".format(i), 0) for i in range(len(eqs))]

    eqs = [str.strip(eq) for eq in eqs]

    eqs = [dolang.parse_string(eq) for eq in eqs]
    es = Sanitizer(variables=model.variables)
    eqs = [es.transform(eq) for eq in eqs]

    eqs = [time_shift(eq, tshift) for eq in eqs]

    eqs = [stringify(eq) for eq in eqs]

    eqs = [str_expression(eq) for eq in eqs]

    targets = [stringify_symbol(e) for e in targets]

    # sanitize defs ( should be )
    defs = dict()

    for k in model.definitions:
        val = model.definitions[k]
        # val = es.transform(dolang.parse_string(val))
        for t in preamble_tshift:
            s = stringify(time_shift(k, t))
            if isinstance(val, str):
                vv = stringify(time_shift(val, t))
            else:
                vv = str(val)
            defs[s] = vv

    preamble = reorder_preamble(defs)

    eqs = dict(zip(targets, eqs))
    ff = FlatFunctionFactory(preamble, eqs, arguments, eq_type)

    return ff



from dolo.numeric.taylor_expansion import CDR
from dolo.numeric.extern.qz import qzordered
from dolo.compiler.derivatives import get_model_derivatives
from typing import List
from numpy import ndarray


def perturb(model, order=1, return_dr=True, steady_state=None, verbose=True):

    from dolo.numeric.processes import IIDProcess

    # here we should check that exogenous don't appear with time t-1 in transition
    # and with time t+1 in arbitrage.
    assert isinstance(model.exogenous, IIDProcess)

    import numpy

    if steady_state is None:
        calibration = model.calibration
    else:
        calibration = steady_state

    [f, g] = get_model_derivatives(model, order=order, calibration=calibration)
    sigma = model.exogenous.Σ

    problem = PerturbationProblem(f, g, sigma)

    pert_sol = state_perturb(problem, verbose=verbose)

    controls_ss = calibration["controls"]
    states_ss = calibration["states"]

    n_s = len(model.symbols["states"])
    n_c = len(model.symbols["controls"])

    if order == 1:
        if return_dr:
            S_bar = numpy.array(states_ss)
            X_bar = numpy.array(controls_ss)
            # add transitions of states to the d.r.

            X_s = pert_sol[0]
            A = g[1][:, :n_s] + numpy.dot(g[1][:, n_s : n_s + n_c], X_s)
            B = g[1][:, n_s + n_c :]
            dr = CDR(S_bar, X_bar, X_s)
            dr.A = A
            dr.B = B
            dr.sigma = sigma
            return dr

        return [controls_ss] + pert_sol

    if order == 2:
        [[X_s, X_ss], [X_tt]] = pert_sol
        X_bar = controls_ss + X_tt / 2
        if return_dr:
            S_bar = states_ss
            S_bar = numpy.array(S_bar)
            X_bar = numpy.array(X_bar)
            dr = CDR(S_bar, X_bar, X_s, X_ss)
            A = g[1][:, :n_s] + numpy.dot(g[1][:, n_s : n_s + n_c], X_s)
            B = g[1][:, n_s + n_c :]
            dr.sigma = sigma
            dr.A = A
            dr.B = B
            return dr
        return [X_bar, X_s, X_ss]

    if order == 3:
        [[X_s, X_ss, X_sss], [X_tt, X_stt]] = pert_sol
        X_bar = controls_ss + X_tt / 2
        X_s = X_s + X_stt / 2
        if return_dr:
            S_bar = states_ss
            dr = CDR(S_bar, X_bar, X_s, X_ss, X_sss)
            dr.sigma = sigma
            return dr
        return [X_bar, X_s, X_ss, X_sss]


class PerturbationProblem:
    def __init__(self, f: List[ndarray], g: List[ndarray], sigma: ndarray):
        self.f = f
        self.g = g
        self.sigma = sigma
        assert len(f) == len(g)

    @property
    def order(self):
        return len(self.f) - 1


approximate_controls = perturb


def state_perturb(problem: PerturbationProblem, verbose=True):
    """Computes a Taylor approximation of decision rules, given the supplied derivatives.

    The original system is assumed to be in the the form:

    .. math::

        E_t f(s_t,x_t,s_{t+1},x_{t+1})

        s_t = g(s_{t-1},x_{t-1}, \\lambda \\epsilon_t)

    where :math:`\\lambda` is a scalar scaling down the risk.  the solution is a function :math:`\\varphi` such that:

    .. math::

        x_t = \\varphi ( s_t, \\sigma )

    The user supplies, a list of derivatives of f and g.

    :param f_fun: list of derivatives of f [order0, order1, order2, ...]
    :param g_fun: list of derivatives of g [order0, order1, order2, ...]
    :param sigma: covariance matrix of :math:`\\epsilon_t`


    Assuming :math:`s_t` ,  :math:`x_t` and :math:`\\epsilon_t` are vectors of size
    :math:`n_s`, :math:`n_x`  and :math:`n_x`  respectively.
    In general the derivative of order :math:`i` of :math:`f`  is a multimensional array of size :math:`n_x \\times (N, ..., N)`
    with :math:`N=2(n_s+n_x)` repeated :math:`i` times (possibly 0).
    Similarly the derivative of order :math:`i` of :math:`g`  is a multidimensional array of size :math:`n_s \\times (M, ..., M)`
    with :math:`M=n_s+n_x+n_2` repeated :math:`i` times (possibly 0).
    """

    import numpy as np
    from numpy.linalg import solve

    approx_order = problem.order  # order of approximation

    [f0, f1] = problem.f[:2]
    [g0, g1] = problem.g[:2]
    sigma = problem.sigma

    n_x = f1.shape[0]  # number of controls
    n_s = f1.shape[1] // 2 - n_x  # number of states
    n_e = g1.shape[1] - n_x - n_s
    n_v = n_s + n_x

    f_s = f1[:, :n_s]
    f_x = f1[:, n_s : n_s + n_x]
    f_snext = f1[:, n_v : n_v + n_s]
    f_xnext = f1[:, n_v + n_s :]

    g_s = g1[:, :n_s]
    g_x = g1[:, n_s : n_s + n_x]
    g_e = g1[:, n_v:]

    A = np.row_stack(
        [
            np.column_stack([np.eye(n_s), np.zeros((n_s, n_x))]),
            np.column_stack([-f_snext, -f_xnext]),
        ]
    )
    B = np.row_stack([np.column_stack([g_s, g_x]), np.column_stack([f_s, f_x])])

    [S, T, Q, Z, eigval] = qzordered(A, B, 1.0 - 1e-8)

    Q = Q.real  # is it really necessary ?
    Z = Z.real

    diag_S = np.diag(S)
    diag_T = np.diag(T)

    tol_geneigvals = 1e-10

    try:
        ok = sum((abs(diag_S) < tol_geneigvals) * (abs(diag_T) < tol_geneigvals)) == 0
        assert ok
    except Exception as e:
        raise GeneralizedEigenvaluesError(diag_S=diag_S, diag_T=diag_T)

    if max(eigval[:n_s]) >= 1 and min(eigval[n_s:]) < 1:
        # BK conditions are met
        pass
    else:
        eigval_s = sorted(eigval, reverse=True)
        ev_a = eigval_s[n_s - 1]
        ev_b = eigval_s[n_s]
        cutoff = (ev_a - ev_b) / 2
        if not ev_a > ev_b:
            raise GeneralizedEigenvaluesSelectionError(
                A=A,
                B=B,
                eigval=eigval,
                cutoff=cutoff,
                diag_S=diag_S,
                diag_T=diag_T,
                n_states=n_s,
            )
        import warnings

        if cutoff > 1:
            warnings.warn("Solution is not convergent.")
        else:
            warnings.warn(
                "There are multiple convergent solutions. The one with the smaller eigenvalues was selected."
            )
        [S, T, Q, Z, eigval] = qzordered(A, B, cutoff)

    Z11 = Z[:n_s, :n_s]
    # Z12 = Z[:n_s, n_s:]
    Z21 = Z[n_s:, :n_s]
    # Z22 = Z[n_s:, n_s:]
    S11 = S[:n_s, :n_s]
    T11 = T[:n_s, :n_s]

    # first order solution
    C = solve(Z11.T, Z21.T).T
    P = np.dot(solve(S11.T, Z11.T).T, solve(Z11.T, T11.T).T)
    Q = g_e

    # if False:
    #     from numpy import dot
    #     test = f_s + f_x @ C + f_snext @ (g_s + g_x @ C) + f_xnext @ C @ (g_s + g_x @ C)
    #     print('Error: ' + str(abs(test).max()))

    if approx_order == 1:
        return [C]

    # second order solution
    from dolo.numeric.tensor import sdot, mdot
    from numpy import dot
    from dolo.numeric.matrix_equations import solve_sylvester

    f2 = problem.f[2]
    g2 = problem.g[2]
    g_ss = g2[:, :n_s, :n_s]
    g_sx = g2[:, :n_s, n_s:n_v]
    g_xx = g2[:, n_s:n_v, n_s:n_v]

    X_s = C

    V1_3 = g_s + dot(g_x, X_s)
    V1 = np.row_stack([np.eye(n_s), X_s, V1_3, X_s @ V1_3])

    K2 = g_ss + 2 * sdot(g_sx, X_s) + mdot(g_xx, X_s, X_s)
    A = f_x + dot(f_snext + dot(f_xnext, X_s), g_x)
    B = f_xnext
    C = V1_3
    D = mdot(f2, V1, V1) + sdot(f_snext + dot(f_xnext, X_s), K2)

    X_ss = solve_sylvester(A, B, C, D)

    #    test = sdot( A, X_ss ) + sdot( B,  mdot(X_ss,V1_3,V1_3) ) + D

    g_ee = g2[:, n_v:, n_v:]

    v = np.row_stack([g_e, dot(X_s, g_e)])

    K_tt = mdot(f2[:, n_v:, n_v:], v, v)
    K_tt += sdot(f_snext + dot(f_xnext, X_s), g_ee)
    K_tt += mdot(sdot(f_xnext, X_ss), g_e, g_e)
    K_tt = np.tensordot(K_tt, sigma, axes=((1, 2), (0, 1)))

    L_tt = f_x + dot(f_snext, g_x) + dot(f_xnext, dot(X_s, g_x) + np.eye(n_x))
    X_tt = solve(L_tt, -K_tt)

    if approx_order == 2:
        return [[X_s, X_ss], [X_tt]]

    # third order solution

    f3 = problem.f[3]
    g3 = problem.g[3]
    g_sss = g3[:, :n_s, :n_s, :n_s]
    g_ssx = g3[:, :n_s, :n_s, n_s:n_v]
    g_sxx = g3[:, :n_s, n_s:n_v, n_s:n_v]
    g_xxx = g3[:, n_s:n_v, n_s:n_v, n_s:n_v]

    V2_3 = K2 + sdot(g_x, X_ss)
    V2 = np.row_stack(
        [np.zeros((n_s, n_s, n_s)), X_ss, V2_3, dot(X_s, V2_3) + mdot(X_ss, V1_3, V1_3)]
    )

    K3 = g_sss + 3 * sdot(g_ssx, X_s) + 3 * mdot(g_sxx, X_s, X_s) + 2 * sdot(g_sx, X_ss)
    K3 += 3 * mdot(g_xx, X_ss, X_s) + mdot(g_xxx, X_s, X_s, X_s)
    L3 = 3 * mdot(X_ss, V1_3, V2_3)

    # A = f_x + dot( f_snext + dot(f_xnext,X_s), g_x) # same as before
    # B = f_xnext # same
    # C = V1_3 # same
    D = (
        mdot(f3, V1, V1, V1)
        + 3 * mdot(f2, V2, V1)
        + sdot(f_snext + dot(f_xnext, X_s), K3)
    )
    D += sdot(f_xnext, L3)

    X_sss = solve_sylvester(A, B, C, D)

    # now doing sigma correction with sigma replaced by l in the subscripts

    g_se = g2[:, :n_s, n_v:]
    g_xe = g2[:, n_s:n_v, n_v:]

    g_see = g3[:, :n_s, n_v:, n_v:]
    g_xee = g3[:, n_s:n_v, n_v:, n_v:]

    W_l = np.row_stack([g_e, dot(X_s, g_e)])

    I_e = np.eye(n_e)

    V_sl = g_se + mdot(g_xe, X_s, np.eye(n_e))

    W_sl = np.row_stack([V_sl, mdot(X_ss, V1_3, g_e) + sdot(X_s, V_sl)])

    K_ee = mdot(f3[:, :, n_v:, n_v:], V1, W_l, W_l)
    K_ee += 2 * mdot(f2[:, n_v:, n_v:], W_sl, W_l)

    # stochastic part of W_ll

    SW_ll = np.row_stack([g_ee, mdot(X_ss, g_e, g_e) + sdot(X_s, g_ee)])

    DW_ll = np.concatenate([X_tt, dot(g_x, X_tt), dot(X_s, sdot(g_x, X_tt)) + X_tt])

    K_ee += mdot(f2[:, :, n_v:], V1, SW_ll)

    K_ = np.tensordot(K_ee, sigma, axes=((2, 3), (0, 1)))

    K_ += mdot(f2[:, :, n_s:], V1, DW_ll)

    def E(vec):
        n = len(vec.shape)
        return np.tensordot(vec, sigma, axes=((n - 2, n - 1), (0, 1)))

    L = sdot(g_sx, X_tt) + mdot(g_xx, X_s, X_tt)

    L += E(g_see + mdot(g_xee, X_s, I_e, I_e))

    M = E(mdot(X_sss, V1_3, g_e, g_e) + 2 * mdot(X_ss, V_sl, g_e))
    M += mdot(X_ss, V1_3, E(g_ee) + sdot(g_x, X_tt))

    A = f_x + dot(f_snext + dot(f_xnext, X_s), g_x)  # same as before
    B = f_xnext  # same
    C = V1_3  # same
    D = K_ + dot(f_snext + dot(f_xnext, X_s), L) + dot(f_xnext, M)

    X_stt = solve_sylvester(A, B, C, D)

    if approx_order == 3:
        # if sigma is None:
        #     return [X_s,X_ss,X_sss]
        # else:
        #     return [[X_s,X_ss,X_sss],[X_tt, X_stt]]
        return [[X_s, X_ss, X_sss], [X_tt, X_stt]]


if __name__ == "__main__":
    from dolo import yaml_import

    model = yaml_import("examples/models/compat/rbc.yaml")
    # model = yaml_import('/home/pablo/Programming/papers/finint/models/integration_B_pert.yaml')

    import time

    t1 = time.time()
    dr = approximate_controls(model, order=2)
    print(dr.X_s)
    # print(dr.X_ss)
    # print(dr.X_sss)
    t2 = time.time()
    print("Elapsed {}".format(t2 - t1))

    t1 = time.time()
    dr = approximate_controls(model, order=2)
    print(dr.X_s)
    # print(dr.X_ss)
    # print(dr.X_sss)
    t2 = time.time()
    print("Elapsed {}".format(t2 - t1))



from functools import reduce
from operator import mul
from quantecon import cartesian
import numpy as np
from numpy import zeros

from typing import TypeVar, Generic, Dict

T = TypeVar("T")
S = TypeVar("S")


def prod(l):
    return reduce(mul, l, 1.0)


from dolo.numeric.misc import mlinspace


class Grid:
    def __mul__(self, rgrid):
        return cat_grids(self, rgrid)

    @property
    def nodes(self):
        return self.__nodes__

    @property
    def n_nodes(self):
        return self.__nodes__.shape[0]

    def node(self, i):
        return self.__nodes__[i, :]


class ProductGrid(Grid, Generic[T, S]):
    def __init__(self, g1: T, g2: S, names=None):
        self.grids = [g1, g2]
        self.names = names

    def __getitem__(self, v):
        return self.grids[self.names.index(v)]

    def __repr__(self):
        return str.join(" × ", [e.__repr__() for e in self.grids])


class EmptyGrid(Grid):

    type = "empty"

    @property
    def nodes(self):
        return None

    @property
    def n_nodes(self):
        return 0

    def node(self, i):
        return None

    def __add__(self, g):
        return g


class PointGrid(Grid):

    type = "point"

    def __init__(self, point):
        self.point = np.array(point)

    @property
    def nodes(self):
        return None

    @property
    def n_nodes(self):

        return 1

    def node(self, i):
        return None


class UnstructuredGrid(Grid):

    type = "unstructured"

    def __init__(self, nodes):
        nodes = np.array(nodes, dtype=float)
        self.min = nodes.min(axis=0)
        self.max = nodes.max(axis=0)
        self.__nodes__ = nodes
        self.d = len(self.min)


class CartesianGrid(Grid):

    pass


class UniformCartesianGrid(CartesianGrid):

    type = "UniformCartesian"

    def __init__(self, min, max, n=[]):

        self.d = len(min)

        # this should be a tuple
        self.min = np.array(min, dtype=float)
        self.max = np.array(max, dtype=float)
        if len(n) == 0:
            self.n = np.zeros(n, dtype=int) + 20
        else:
            self.n = np.array(n, dtype=int)

        # this should be done only on request.
        self.__nodes__ = mlinspace(self.min, self.max, self.n)

    # def node(i:)
    # pass

    def __add__(self, g):

        if not isinstance(g, UniformCartesianGrid):
            raise Exception("Not implemented.")

        n = np.array(tuple(self.n) + tuple(g.n))
        min = np.array(tuple(self.min) + tuple(self.min))
        max = np.array(tuple(self.max) + tuple(self.max))

        return UniformCartesianGrid(min, max, n)

    def __numba_repr__(self):
        return tuple([(self.min[i], self.max[i], self.n[i]) for i in range(self.d)])


class NonUniformCartesianGrid(CartesianGrid):

    type = "NonUniformCartesian"

    def __init__(self, list_of_nodes):
        list_of_nodes = [np.array(l) for l in list_of_nodes]
        self.min = [min(l) for l in list_of_nodes]
        self.max = [max(l) for l in list_of_nodes]
        self.n = np.array([(len(e)) for e in list_of_nodes])
        # this should be done only on request.
        self.__nodes__ = cartesian(list_of_nodes)
        self.list_of_nodes = list_of_nodes  # think of a better name

    def __add__(self, g):

        if not isinstance(g, NonUniformCartesianGrid):
            raise Exception("Not implemented.")
        return NonUniformCartesianGrid(self.list_of_nodes + g.list_of_nodes)

    def __numba_repr__(self):
        return tuple([np.array(e) for e in self.list_of_nodes])


class SmolyakGrid(Grid):

    type = "Smolyak"

    def __init__(self, min, max, mu=2):

        from interpolation.smolyak import SmolyakGrid as ISmolyakGrid

        min = np.array(min)
        max = np.array(max)
        self.min = min
        self.max = max
        self.mu = mu
        d = len(min)
        sg = ISmolyakGrid(d, mu, lb=min, ub=max)
        self.sg = sg
        self.d = d
        self.__nodes__ = sg.grid


def cat_grids(grid_1, grid_2):

    if isinstance(grid_1, EmptyGrid):
        return grid_2
    if isinstance(grid_1, CartesianGrid) and isinstance(grid_2, CartesianGrid):
        min = np.concatenate([grid_1.min, grid_2.min])
        max = np.concatenate([grid_1.max, grid_2.max])
        n = np.concatenate([grid_1.n, grid_2.n])
        return CartesianGrid(min, max, n)
    else:
        raise Exception("Not Implemented.")


# compat
def node(grid, i):
    return grid.node(i)


def nodes(grid):
    return grid.nodes


def n_nodes(grid):
    return grid.n_nodes


if __name__ == "__main__":

    print("Cartsian Grid")
    grid = CartesianGrid([0.1, 0.3], [9, 0.4], [50, 10])
    print(grid.nodes)
    print(nodes(grid))

    print("UnstructuredGrid")
    ugrid = UnstructuredGrid([[0.1, 0.3], [9, 0.4], [50, 10]])
    print(nodes(ugrid))
    print(node(ugrid, 0))
    print(n_nodes(ugrid))

    print("Non Uniform CartesianGrid")
    ugrid = NonUniformCartesianGrid([[0.1, 0.3], [9, 0.4], [50, 10]])
    print(nodes(ugrid))
    print(node(ugrid, 0))
    print(n_nodes(ugrid))

    print("Smolyak Grid")
    sg = SmolyakGrid([0.1, 0.2], [1.0, 2.0], 2)
    print(nodes(sg))
    print(node(sg, 1))
    print(n_nodes(sg))



import numpy as np
import pandas as pd
from numpy import array, atleast_2d, linspace, zeros
from scipy.optimize import root

from dolo.numeric.optimize.ncpsolve import ncpsolve


def _shocks_to_epsilons(model, shocks, T):
    """
    Helper function to support input argument `shocks` being one of many
    different data types. Will always return a `T, n_e` matrix.
    """
    n_e = len(model.calibration["exogenous"])

    # if we have a DataFrame, convert it to a dict and rely on the method below
    if isinstance(shocks, pd.DataFrame):
        shocks = {k: shocks[k].tolist() for k in shocks.columns}

    # handle case where shocks might be a dict. Be careful to handle case where
    # value arrays are not the same length
    if isinstance(shocks, dict):
        epsilons = np.zeros((T + 1, n_e))
        for (i, k) in enumerate(model.symbols["exogenous"]):
            if k in shocks:
                this_shock = shocks[k]
                epsilons[: len(this_shock), i] = this_shock
                epsilons[len(this_shock) :, i] = this_shock[-1]
            else:
                # otherwise set to value in calibration
                epsilons[:, i] = model.calibration["exogenous"][i]

        return epsilons

    # read from calibration if not given
    if shocks is None:
        shocks = model.calibration["exogenous"]

    # now we just assume that shocks is array-like and try using the output of
    # np.asarray(shocks)
    shocks = np.asarray(shocks)
    shocks = shocks.reshape((-1, n_e))

    # until last period, exogenous shock takes its last value
    epsilons = np.zeros((T + 1, n_e))
    epsilons[: (shocks.shape[0] - 1), :] = shocks[1:, :]
    epsilons[(shocks.shape[0] - 1) :, :] = shocks[-1:, :]

    return epsilons


def deterministic_solve(
    model,
    exogenous=None,
    s0=None,
    m0=None,
    T=100,
    ignore_constraints=False,
    maxit=100,
    initial_guess=None,
    verbose=True,
    solver="ncpsolve",
    keep_steady_state=False,
    s1=None,  # deprecated
    shocks=None,  # deprecated
    tol=1e-6,
):
    """
    Computes a perfect foresight simulation using a stacked-time algorithm.

    Typical simulation exercises are:
    - start from an out-of-equilibrium exogenous and/or endogenous state: specify `s0` and or `m0`. Missing values are taken from the calibration (`model.calibration`).
    - specify an exogenous path for shocks `exogenous`. Initial exogenous state `m0` is then first value of exogenous values. Economy is supposed to have been at the equilibrium for $t<0$, which pins
    down initial endogenous state `s0`. `x0` is a jump variable.

    If $s0$ is not specified it is then set
    equal to the steady-state consistent with the first value

    The initial state is specified either by providing a series of exogenous
    shocks and assuming the model is initially in equilibrium with the first
    value of the shock, or by specifying an initial value for the states.

    Parameters
    ----------
    model : Model
        Model to be solved
    exogenous : array-like, dict, or pandas.DataFrame
        A specification for the path of exogenous variables (aka shocks). Can be any of the
        following (note by "declaration order" below we mean the order
        of `model.symbols["exogenous"]`):

        - A 1d numpy array-like specifying a time series for a single
          exogenous variable, or all exogenous variables stacked into a single array.
        - A 2d numpy array where each column specifies the time series
          for one of the shocks in declaration order. This must be an
          `N` by number of shocks 2d array.
        - A dict where keys are strings found in
          `model.symbols["exogenous"]` and values are a time series of
          values for that shock. For exogenous variables that do not appear in
          this dict, the shock is set to the calibrated value. Note
          that this interface is the most flexible as it allows the user
          to pass values for only a subset of the model shocks and it
          allows the passed time series to be of different lengths.
        - A DataFrame where columns map shock names into time series.
          The same assumptions and behavior that are used in the dict
          case apply here

        If nothing is given here, `exogenous` is set equal to the
        calibrated values found in `model.calibration["exogenous"]` for
        all periods.

        If the length of any time-series in shocks is less than `T`
        (see below) it is assumed that that particular shock will
        remain at the final given value for the duration of the
        simulation.
    s0 : None or ndarray or dict
        If vector with the value of initial states
        If an exogenous timeseries is given for exogenous shocks, `s0` will be computed as the steady-state value that is consistent with its first value.

    T : int
        horizon for the perfect foresight simulation
    maxit : int
        maximum number of iteration for the nonlinear solver
    verbose : boolean
        if True, the solver displays iterations
    tol : float
        stopping criterium for the nonlinear solver
    ignore_constraints : bool
        if True, complementarity constraints are ignored.
    keep_steady_state : bool
        if True, initial steady-states and steady-controls are appended to the simulation with date -1.
    Returns
    -------
    pandas dataframe
        a dataframe with T+1 observations of the model variables along the
        simulation (states, controls, auxiliaries). The  simulation should return to a steady-state
        consistent with the last specified value of the exogenous shocks.

    """

    if shocks is not None:
        import warnings

        warnings.warn("`shocks` argument is deprecated. Use `exogenous` instead.")
        exogenous = shocks

    if s1 is not None:
        import warnings

        warnings.warn("`s1` argument is deprecated. Use `s0` instead.")
        s0 = s1

    # definitions
    n_s = len(model.calibration["states"])
    n_x = len(model.calibration["controls"])
    p = model.calibration["parameters"]

    if exogenous is not None:
        epsilons = _shocks_to_epsilons(model, exogenous, T)
        m0 = epsilons[0, :]
        # get initial steady-state
        from dolo.algos.steady_state import find_steady_state

        start_state = find_steady_state(model, m=m0)
        s0 = start_state["states"]
        x0 = start_state["controls"]
        m1 = epsilons[1, :]
        s1 = model.functions["transition"](m0, s0, x0, m1, p)
    else:
        if s0 is None:
            s0 = model.calibration["states"]
        if m0 is None:
            m0 = model.calibration["exogenous"]
        # if m0 is None:
        #     m0 = np.zeros(len(model.symbols['exogenous']))
        # we should probably do something here with the nature of the exogenous process if specified
        # i.e. compute nonlinear irf
        epsilons = _shocks_to_epsilons(model, exogenous, T)
        x0 = model.calibration["controls"]
        m1 = epsilons[1, :]
        s1 = model.functions["transition"](m0, s0, x0, m1, p)
        s1 = np.array(s1)

    x1_g = model.calibration["controls"]  # we can do better here
    sT_g = model.calibration["states"]  # we can do better here
    xT_g = model.calibration["controls"]  # we can do better here

    if initial_guess is None:
        start = np.concatenate([s1, x1_g])
        final = np.concatenate([sT_g, xT_g])
        initial_guess = np.row_stack(
            [start * (1 - l) + final * l for l in linspace(0.0, 1.0, T + 1)]
        )

    else:
        if isinstance(initial_guess, pd.DataFrame):
            initial_guess = np.array(
                initial_guess[model.symbols["states"] + model.symbols["controls"]]
            )
        initial_guess = initial_guess[1:, :]
        initial_guess = initial_guess[:, : n_s + n_x]

    sh = initial_guess.shape

    if model.x_bounds and not ignore_constraints:
        initial_states = initial_guess[:, :n_s]
        [lb, ub] = [u(epsilons[:, :], initial_states, p) for u in model.x_bounds]
        lower_bound = initial_guess * 0 - np.inf
        lower_bound[:, n_s:] = lb
        upper_bound = initial_guess * 0 + np.inf
        upper_bound[:, n_s:] = ub
        test1 = max(lb.max(axis=0) - lb.min(axis=0))
        test2 = max(ub.max(axis=0) - ub.min(axis=0))
        if test1 > 0.00000001 or test2 > 0.00000001:
            msg = "Not implemented: perfect foresight solution requires that "
            msg += "controls have constant bounds."
            raise Exception(msg)
    else:
        ignore_constraints = True
        lower_bound = None
        upper_bound = None

    det_residual(model, initial_guess, s1, xT_g, epsilons)

    if not ignore_constraints:

        def ff(vec):
            return det_residual(
                model, vec.reshape(sh), s1, xT_g, epsilons, jactype="sparse"
            )

        v0 = initial_guess.ravel()
        if solver == "ncpsolve":
            sol, nit = ncpsolve(
                ff,
                lower_bound.ravel(),
                upper_bound.ravel(),
                initial_guess.ravel(),
                verbose=verbose,
                maxit=maxit,
                tol=tol,
                jactype="sparse",
            )
        else:
            from dolo.numeric.extern.lmmcp import lmmcp

            sol = lmmcp(
                lambda u: ff(u)[0],
                lambda u: ff(u)[1].todense(),
                lower_bound.ravel(),
                upper_bound.ravel(),
                initial_guess.ravel(),
                verbose=verbose,
            )
            nit = -1

        sol = sol.reshape(sh)

    else:

        def ff(vec):
            ll = det_residual(model, vec.reshape(sh), s1, xT_g, epsilons, diff=True)
            return ll

        v0 = initial_guess.ravel()
        # from scipy.optimize import root
        # sol = root(ff, v0, jac=True)
        # sol = sol.x.reshape(sh)
        from dolo.numeric.optimize.newton import newton

        sol, nit = newton(ff, v0, jactype="sparse")
        sol = sol.reshape(sh)

    # sol = sol[:-1, :]

    if (exogenous is not None) and keep_steady_state:
        sx = np.concatenate([s0, x0])
        sol = np.concatenate([sx[None, :], sol], axis=0)
        epsilons = np.concatenate([epsilons[:1:], epsilons], axis=0)
        index = range(-1, T + 1)
    else:
        index = range(0, T + 1)
    # epsilons = np.concatenate([epsilons[:1,:], epsilons], axis=0)

    if "auxiliary" in model.functions:
        colnames = (
            model.symbols["states"]
            + model.symbols["controls"]
            + model.symbols["auxiliaries"]
        )
        # compute auxiliaries
        y = model.functions["auxiliary"](epsilons, sol[:, :n_s], sol[:, n_s:], p)
        sol = np.column_stack([sol, y])
    else:
        colnames = model.symbols["states"] + model.symbols["controls"]

    sol = np.column_stack([sol, epsilons])
    colnames = colnames + model.symbols["exogenous"]

    ts = pd.DataFrame(sol, columns=colnames, index=index)
    return ts


def det_residual(model, guess, start, final, shocks, diff=True, jactype="sparse"):
    """
    Computes the residuals, the derivatives of the stacked-time system.
    :param model: an fga model
    :param guess: the guess for the simulated values. An `(n_s.n_x) x N` array,
                  where n_s is the number of states,
    n_x the number of controls, and `N` the length of the simulation.
    :param start: initial boundary condition (initial value of the states)
    :param final: final boundary condition (last value of the controls)
    :param shocks: values for the exogenous shocks
    :param diff: if True, the derivatives are computes
    :return: a list with two elements:
        - an `(n_s.n_x) x N` array with the residuals of the system
        - a `(n_s.n_x) x N x (n_s.n_x) x N` array representing the jacobian of
             the system
    """

    # TODO: compute a sparse derivative and ensure the solvers can deal with it

    n_s = len(model.symbols["states"])
    n_x = len(model.symbols["controls"])

    # n_e = len(model.symbols['shocks'])
    N = guess.shape[0]

    p = model.calibration["parameters"]

    f = model.functions["arbitrage"]
    g = model.functions["transition"]

    vec = guess[:-1, :]
    vec_f = guess[1:, :]

    s = vec[:, :n_s]
    x = vec[:, n_s:]
    S = vec_f[:, :n_s]
    X = vec_f[:, n_s:]

    m = shocks[:-1, :]
    M = shocks[1:, :]

    if diff:
        SS, SS_m, SS_s, SS_x, SS_M = g(m, s, x, M, p, diff=True)
        R, R_m, R_s, R_x, R_M, R_S, R_X = f(m, s, x, M, S, X, p, diff=True)
    else:
        SS = g(m, s, x, M, p)
        R = f(m, s, x, M, S, X, p)

    res_s = SS - S
    res_x = R

    res = np.zeros((N, n_s + n_x))

    res[1:, :n_s] = res_s
    res[:-1, n_s:] = res_x

    res[0, :n_s] = -(guess[0, :n_s] - start)
    res[-1, n_s:] = -(guess[-1, n_s:] - guess[-2, n_s:])

    if not diff:
        return res
    else:

        sparse_jac = False
        if not sparse_jac:

            # we compute the derivative matrix
            res_s_s = SS_s
            res_s_x = SS_x

            # next block is probably very inefficient
            jac = np.zeros((N, n_s + n_x, N, n_s + n_x))
            for i in range(N - 1):
                jac[i, n_s:, i, :n_s] = R_s[i, :, :]
                jac[i, n_s:, i, n_s:] = R_x[i, :, :]
                jac[i, n_s:, i + 1, :n_s] = R_S[i, :, :]
                jac[i, n_s:, i + 1, n_s:] = R_X[i, :, :]
                jac[i + 1, :n_s, i, :n_s] = SS_s[i, :, :]
                jac[i + 1, :n_s, i, n_s:] = SS_x[i, :, :]
                jac[i + 1, :n_s, i + 1, :n_s] = -np.eye(n_s)
                # jac[i,n_s:,i,:n_s] = R_s[i,:,:]
                # jac[i,n_s:,i,n_s:] = R_x[i,:,:]
                # jac[i+1,n_s:,i,:n_s] = R_S[i,:,:]
                # jac[i+1,n_s:,i,n_s:] = R_X[i,:,:]
                # jac[i,:n_s,i+1,:n_s] = SS_s[i,:,:]
                # jac[i,:n_s,i+1,n_s:] = SS_x[i,:,:]
                # jac[i+1,:n_s,i+1,:n_s] = -np.eye(n_s)
            jac[0, :n_s, 0, :n_s] = -np.eye(n_s)
            jac[-1, n_s:, -1, n_s:] = -np.eye(n_x)
            jac[-1, n_s:, -2, n_s:] = +np.eye(n_x)
            nn = jac.shape[0] * jac.shape[1]
            res = res.ravel()
            jac = jac.reshape((nn, nn))

        if jactype == "sparse":
            from scipy.sparse import csc_matrix, csr_matrix

            jac = csc_matrix(jac)
            # scipy bug ? I don't get the same with csr

        return [res, jac]



import sys

is_python_3 = sys.version_info >= (3, 0)


def read_file_or_url(url):

    if "http" in url:
        if is_python_3:
            import urllib.request

            txt = urllib.request.urlopen(url).read()
            txt = txt.decode("utf8")  # not very robust
        else:
            import urllib2

            txt = urllib2.urlopen(url).read()
    else:
        # must be a file
        with open(url, encoding="utf8") as f:
            txt = f.read()

    return txt


def pcat(filename, target="ipython"):

    code = read_file_or_url(filename)

    try:
        assert target == "ipython"
        from IPython import get_ipython

        ipython = get_ipython()

        ipython.magic(f"pycat {filename}")
        return

    except Exception as e:
        print(e)
        pass

    from pygments.lexers import get_lexer_for_filename
    from pygments.formatters import TerminalFormatter
    from pygments import highlight

    lexer = get_lexer_for_filename(filename, stripall=True)
    formatter = TerminalFormatter()
    output = highlight(code, lexer, formatter)
    print(output)



from dolang.symbolic import sanitize, parse_string, str_expression
from dolang.language import eval_data
from dolang.symbolic import str_expression

import copy


class SymbolicModel:
    def __init__(self, data):

        self.data = data

    @property
    def symbols(self):

        if self.__symbols__ is None:

            from .misc import LoosyDict, equivalent_symbols
            from dolang.symbolic import remove_timing, parse_string, str_expression

            symbols = LoosyDict(equivalences=equivalent_symbols)
            for sg in self.data["symbols"].keys():
                symbols[sg] = [s.value for s in self.data["symbols"][sg]]

            self.__symbols__ = symbols

            # the following call adds auxiliaries (tricky, isn't it?)
            self.definitions

        return self.__symbols__

    @property
    def variables(self):
        if self.__variables__ is None:

            self.__variables__ = sum(
                [self.symbols[e] for e in self.symbols.keys() if e != "parameters"], []
            )

        return self.__variables__

    @property
    def equations(self):
        import yaml.nodes

        if self.__equations__ is None:

            vars = self.variables + [*self.definitions.keys()]

            d = dict()
            for g, v in self.data["equations"].items():

                # new style
                if isinstance(v, yaml.nodes.ScalarNode):
                    assert v.style == "|"
                    if g in ("arbitrage",):
                        start = "complementarity_block"
                    else:
                        start = "assignment_block"
                    eqs = parse_string(v, start=start)
                    eqs = sanitize(eqs, variables=vars)
                    eq_list = eqs.children
                # old style
                else:
                    eq_list = []
                    for eq_string in v:
                        start = "equation"  # it should be assignment
                        eq = parse_string(eq_string, start=start)
                        eq = sanitize(eq, variables=vars)
                        eq_list.append(eq)

                if g in ("arbitrage",):
                    ll = []  # List[str]
                    ll_lb = []  # List[str]
                    ll_ub = []  # List[str]
                    with_complementarity = False
                    for i, eq in enumerate(eq_list):
                        if eq.data == "double_complementarity":
                            v = eq.children[1].children[1].children[0].children[0].value
                            t = int(
                                eq.children[1].children[1].children[1].children[0].value
                            )
                            expected = (
                                self.symbols["controls"][i],
                                0,
                            )  # TODO raise nice error message
                            if (v, t) != expected:
                                raise Exception(
                                    f"Incorrect variable in complementarity: expected {expected}. Found {(v,t)}"
                                )
                            ll_lb.append(str_expression(eq.children[1].children[0]))
                            ll_ub.append(str_expression(eq.children[1].children[2]))
                            eq = eq.children[0]
                            with_complementarity = True
                        else:
                            ll_lb.append("-inf")
                            ll_ub.append("inf")
                        from dolang.symbolic import list_symbols

                        # syms = list_symbols(eq)
                        ll.append(str_expression(eq))
                    d[g] = ll
                    if with_complementarity:
                        d[g + "_lb"] = ll_lb
                        d[g + "_ub"] = ll_ub
                else:
                    # TODO: we should check here that equations are well specified
                    d[g] = [str_expression(e) for e in eq_list]

            # if "controls_lb" not in d:
            #     for ind, g in enumerate(("controls_lb", "controls_ub")):
            #         eqs = []
            #         for i, eq in enumerate(d['arbitrage']):
            #             if "⟂" not in eq:
            #                 if ind == 0:
            #                     eq = "-inf"
            #                 else:
            #                     eq = "inf"
            #             else:
            #                 comp = eq.split("⟂")[1].strip()
            #                 v = self.symbols["controls"][i]
            #                 eq = decode_complementarity(comp, v+"[t]")[ind]
            #             eqs.append(eq)
            #         d[g] = eqs

            self.__equations__ = d

        return self.__equations__

    @property
    def definitions(self):

        from yaml import ScalarNode

        if self.__definitions__ is None:

            # at this stage, basic_symbols doesn't contain auxiliaries
            basic_symbols = self.symbols
            vars = sum(
                [basic_symbols[k] for k in basic_symbols.keys() if k != "parameters"],
                [],
            )

            # # auxiliaries = [remove_timing(parse_string(k)) for k in self.data.get('definitions', {})]
            # # auxiliaries = [str_expression(e) for e in auxiliaries]
            # # symbols['auxiliaries'] = auxiliaries

            if "definitions" not in self.data:
                self.__definitions__ = {}
                # self.__symbols__['auxiliaries'] = []

            elif isinstance(self.data["definitions"], ScalarNode):

                definitions = {}

                # new-style
                from lark import Token

                def_block_tree = parse_string(
                    self.data["definitions"], start="assignment_block"
                )
                def_block_tree = sanitize(
                    def_block_tree
                )  # just to replace (v,) by (v,0) # TODO: remove

                auxiliaries = []
                for eq_tree in def_block_tree.children:
                    lhs, rhs = eq_tree.children
                    tok_name: Token = lhs.children[0].children[0]
                    tok_date: Token = lhs.children[1].children[0]
                    name = tok_name.value
                    date = int(tok_date.value)
                    if name in vars:
                        raise Exception(
                            f"definitions:{tok_name.line}:{tok_name.column}: Auxiliary variable '{name}'' already defined."
                        )
                    if date != 0:
                        raise Exception(
                            f"definitions:{tok_name.line}:{tok_name.column}: Auxiliary variable '{name}' must be defined at date 't'."
                        )
                    # here we could check some stuff
                    from dolang import list_symbols

                    syms = list_symbols(rhs)
                    for p in syms.parameters:
                        if p in vars:
                            raise Exception(
                                f"definitions:{tok_name.line}: Symbol '{p}' is defined as a variable. Can't appear as a parameter."
                            )
                        if p not in self.symbols["parameters"]:
                            raise Exception(
                                f"definitions:{tok_name.line}: Paremeter '{p}' must be defined as a model symbol."
                            )
                    for v in syms.variables:
                        if v[0] not in vars:
                            raise Exception(
                                f"definitions:{tok_name.line}: Variable '{v[0]}[t]' is not defined."
                            )
                    auxiliaries.append(name)
                    vars.append(name)

                    definitions[str_expression(lhs)] = str_expression(rhs)

                self.__symbols__["auxiliaries"] = auxiliaries
                self.__definitions__ = definitions

            else:

                # old style
                from dolang.symbolic import remove_timing

                auxiliaries = [
                    remove_timing(parse_string(k))
                    for k in self.data.get("definitions", {})
                ]
                auxiliaries = [str_expression(e) for e in auxiliaries]
                self.__symbols__["auxiliaries"] = auxiliaries
                vars = self.variables
                auxs = []

                definitions = self.data["definitions"]
                d = dict()
                for i in range(len(definitions.value)):

                    kk = definitions.value[i][0]
                    if self.__compat__:
                        k = parse_string(kk.value)
                        if k.data == "symbol":
                            # TODO: warn that definitions should be timed
                            from dolang.grammar import create_variable

                            k = create_variable(k.children[0].value, 0)
                    else:
                        k = parse_string(kk.value, start="variable")
                    k = sanitize(k, variables=vars)

                    assert k.children[1].children[0].value == "0"

                    vv = definitions.value[i][1]
                    v = parse_string(vv, start="formula")
                    v = sanitize(v, variables=vars)
                    v = str_expression(v)

                    key = str_expression(k)
                    vars.append(key)
                    d[key] = v
                    auxs.append(remove_timing(key))

                self.__symbols__["auxiliaries"] = auxs
                self.__definitions__ = d

        return self.__definitions__

    @property
    def name(self):
        try:
            self.data["name"].value
        except Exception as e:
            return "Anonymous"

    @property
    def infos(self):
        infos = {
            "name": self.name,
            "filename": self.data.get("filename", "<string>"),
            "type": "dtcc",
        }
        return infos

    @property
    def options(self):
        return self.data["options"]

    def get_calibration(self):

        # if self.__calibration__ is None:

        from dolang.symbolic import remove_timing

        import copy

        symbols = self.symbols
        calibration = dict()
        for k, v in self.data.get("calibration", {}).items():
            if v.tag == "tag:yaml.org,2002:str":

                expr = parse_string(v)
                expr = remove_timing(expr)
                expr = str_expression(expr)
            else:
                expr = float(v.value)
            kk = remove_timing(parse_string(k))
            kk = str_expression(kk)

            calibration[kk] = expr

        definitions = self.definitions

        initial_values = {
            "exogenous": 0,
            "expectations": 0,
            "values": 0,
            "controls": float("nan"),
            "states": float("nan"),
        }

        # variables defined by a model equation default to using these definitions
        initialized_from_model = {
            "values": "value",
            "expectations": "expectation",
            "direct_responses": "direct_response",
        }
        for k, v in definitions.items():
            kk = remove_timing(k)
            if kk not in calibration:
                if isinstance(v, str):
                    vv = remove_timing(v)
                else:
                    vv = v
                calibration[kk] = vv

        for symbol_group in symbols:
            if symbol_group not in initialized_from_model.keys():
                if symbol_group in initial_values:
                    default = initial_values[symbol_group]
                else:
                    default = float("nan")
                for s in symbols[symbol_group]:
                    if s not in calibration:
                        calibration[s] = default

        from dolang.triangular_solver import solve_triangular_system

        return solve_triangular_system(calibration)

    #     self.__calibration__ =  solve_triangular_system(calibration)

    # return self.__calibration__

    def get_domain(self):

        calibration = self.get_calibration()
        states = self.symbols["states"]

        sdomain = self.data.get("domain", {})
        for k in sdomain.keys():
            if k not in states:
                sdomain.pop(k)

        # backward compatibility
        if len(sdomain) == 0 and len(states) > 0:
            try:
                import warnings

                min = get_address(self.data, ["options:grid:a", "options:grid:min"])
                max = get_address(self.data, ["options:grid:b", "options:grid:max"])
                for i, s in enumerate(states):
                    sdomain[s] = [min[i], max[i]]
                # shall we raise a warning for deprecated syntax ?
            except Exception as e:
                pass

        if len(sdomain) == 0:
            return None

        if len(sdomain) < len(states):
            missing = [s for s in states if s not in sdomain]
            raise Exception(
                "Missing domain for states: {}.".format(str.join(", ", missing))
            )

        from dolo.compiler.objects import CartesianDomain
        from dolang.language import eval_data

        sdomain = eval_data(sdomain, calibration)

        domain = CartesianDomain(**sdomain)

        return domain

    def get_exogenous(self):

        if "exogenous" not in self.data:
            return {}

        exo = self.data["exogenous"]
        calibration = self.get_calibration()
        from dolang.language import eval_data

        exogenous = eval_data(exo, calibration)

        from dolo.numeric.processes import ProductProcess, Process

        if isinstance(exogenous, Process):
            # old style
            return exogenous
        elif isinstance(exo, list):
            # old style (2)
            return ProductProcess(*exogenous)
        else:
            # new style
            syms = self.symbols["exogenous"]
            # first we check that shocks are defined in the right order
            ssyms = []
            for k in exo.keys():
                vars = [v.strip() for v in k.split(",")]
                ssyms.append(vars)
            ssyms = tuple(sum(ssyms, []))
            if tuple(syms) != ssyms:
                from dolang.language import ModelError

                lc = exo.lc
                raise ModelError(
                    f"{lc.line}:{lc.col}: 'exogenous' section. Shocks specification must match declaration order. Found {ssyms}. Expected{tuple(syms)}"
                )

            return ProductProcess(*exogenous.values())

    @property
    def endo_grid(self):

        # determine bounds:
        domain = self.get_domain()
        min = domain.min
        max = domain.max

        options = self.data.get("options", {})

        # determine grid_type
        grid_type = get_type(options.get("grid"))
        if grid_type is None:
            grid_type = get_address(
                self.data, ["options:grid:type", "options:grid_type"]
            )
        if grid_type is None:
            raise Exception('Missing grid geometry ("options:grid:type")')

        args = {"min": min, "max": max}
        if grid_type.lower() in ("cartesian", "cartesiangrid"):
            from dolo.numeric.grids import UniformCartesianGrid

            orders = get_address(self.data, ["options:grid:n", "options:grid:orders"])
            if orders is None:
                orders = [20] * len(min)
            grid = UniformCartesianGrid(min=min, max=max, n=orders)
        elif grid_type.lower() in ("nonuniformcartesian", "nonuniformcartesiangrid"):
            from dolang.language import eval_data
            from dolo.numeric.grids import NonUniformCartesianGrid

            calibration = self.get_calibration()
            nodes = [eval_data(e, calibration) for e in self.data["options"]["grid"]]
            # each element of nodes should be a vector
            return NonUniformCartesianGrid(nodes)
        elif grid_type.lower() in ("smolyak", "smolyakgrid"):
            from dolo.numeric.grids import SmolyakGrid

            mu = get_address(self.data, ["options:grid:mu"])
            if mu is None:
                mu = 2
            grid = SmolyakGrid(min=min, max=max, mu=mu)
        else:
            raise Exception("Unknown grid type.")

        return grid


def get_type(d):
    try:
        s = d.tag
        return s.strip("!")
    except:
        v = d.get("type")
        return v


def get_address(data, address, default=None):

    if isinstance(address, list):
        found = [get_address(data, e, None) for e in address]
        found = [f for f in found if f is not None]
        if len(found) > 0:
            return found[0]
        else:
            return default
    fields = str.split(address, ":")
    while len(fields) > 0:
        data = data.get(fields[0])
        fields = fields[1:]
        if data is None:
            return default
    return eval_data(data)


import re

regex = re.compile("(.*)<=(.*)<=(.*)")


def decode_complementarity(comp, control):
    """
    # comp can be either:
    - None
    - "a<=expr" where a is a controls
    - "expr<=a" where a is a control
    - "expr1<=a<=expr2"
    """

    try:
        res = regex.match(comp).groups()
    except:
        raise Exception("Unable to parse complementarity condition '{}'".format(comp))

    res = [r.strip() for r in res]
    if res[1] != control:
        msg = "Complementarity condition '{}' incorrect. Expected {} instead of {}.".format(
            comp, control, res[1]
        )
        raise Exception(msg)

    return [res[0], res[2]]


class Model(SymbolicModel):
    """Model Object"""

    def __init__(self, data, check=True, compat=True):

        self.__compat__ = True

        super().__init__(data)

        self.model_type = "dtcc"
        self.__functions__ = None
        # self.__compile_functions__()
        self.set_changed(all="True")

        if check:
            self.symbols
            self.definitions
            self.calibration
            self.domain
            self.exogenous
            self.x_bounds
            self.functions

    def set_changed(self, all=False):
        self.__domain__ = None
        self.__exogenous__ = None
        self.__calibration__ = None
        if all:
            self.__symbols__ = None
            self.__definitions__ = None
            self.__variables__ = None
            self.__equations__ = None

    def set_calibration(self, *pargs, **kwargs):
        if len(pargs) == 1:
            self.set_calibration(**pargs[0])
        self.set_changed()
        self.data["calibration"].update(kwargs)

    @property
    def calibration(self):
        if self.__calibration__ is None:
            calibration_dict = super().get_calibration()
            from dolo.compiler.misc import CalibrationDict, calibration_to_vector

            calib = calibration_to_vector(self.symbols, calibration_dict)
            self.__calibration__ = CalibrationDict(self.symbols, calib)  #
        return self.__calibration__

    @property
    def exogenous(self):
        if self.__exogenous__ is None:
            self.__exogenous__ = super(self.__class__, self).get_exogenous()
        return self.__exogenous__

    @property
    def domain(self):
        if self.__domain__ is None:
            self.__domain__ = super().get_domain()
        return self.__domain__

    def discretize(self, grid_options=None, dprocess_options={}):

        dprocess = self.exogenous.discretize(**dprocess_options)

        if grid_options is None:
            endo_grid = self.endo_grid
        else:
            endo_grid = self.domain.discretize(**grid_options)

        from dolo.numeric.grids import ProductGrid

        grid = ProductGrid(dprocess.grid, endo_grid, names=["exo", "endo"])
        return [grid, dprocess]

    def __compile_functions__(self):

        from dolang.function_compiler import make_method_from_factory

        from dolang.vectorize import standard_function
        from dolo.compiler.factories import get_factory
        from .misc import LoosyDict

        equivalent_function_names = {
            "equilibrium": "arbitrage",
            "optimality": "arbitrage",
        }
        functions = LoosyDict(equivalences=equivalent_function_names)
        original_functions = {}
        original_gufunctions = {}

        funnames = [*self.equations.keys()]
        if len(self.definitions) > 0:
            funnames = funnames + ["auxiliary"]

        import dolo.config

        debug = dolo.config.debug

        for funname in funnames:

            fff = get_factory(self, funname)
            fun, gufun = make_method_from_factory(fff, vectorize=True, debug=debug)
            n_output = len(fff.content)
            functions[funname] = standard_function(gufun, n_output)
            original_gufunctions[funname] = gufun  # basic gufun function
            original_functions[funname] = fun  # basic numba fun

        self.__original_functions__ = original_functions
        self.__original_gufunctions__ = original_gufunctions
        self.__functions__ = functions

    @property
    def functions(self):
        if self.__functions__ is None:
            self.__compile_functions__()
        return self.__functions__

    def __str__(self):

        from dolo.misc.termcolor import colored
        from numpy import zeros

        s = """
        Model:
        ------
        name: "{name}"
        type: "{type}"
        file: "{filename}\n""".format(
            **self.infos
        )

        ss = "\nEquations:\n----------\n\n"
        res = self.residuals()
        res.update({"definitions": zeros(1)})

        equations = self.equations.copy()
        definitions = self.definitions
        tmp = []
        for deftype in definitions:
            tmp.append(deftype + " = " + definitions[deftype])
        definitions = {"definitions": tmp}
        equations.update(definitions)
        # for eqgroup, eqlist in self.symbolic.equations.items():
        for eqgroup in res.keys():
            if eqgroup == "auxiliary":
                continue
            if eqgroup == "definitions":
                eqlist = equations[eqgroup]
                # Update the residuals section with the right number of empty
                # values. Note: adding 'zeros' was easiest (rather than empty
                # cells), since other variable types have  arrays of zeros.
                res.update({"definitions": [None for i in range(len(eqlist))]})
            else:
                eqlist = equations[eqgroup]
            ss += "{}\n".format(eqgroup)
            for i, eq in enumerate(eqlist):
                val = res[eqgroup][i]
                if val is None:
                    ss += " {eqn:2} : {eqs}\n".format(eqn=str(i + 1), eqs=eq)
                else:
                    if abs(val) < 1e-8:
                        val = 0
                    vals = "{:.4f}".format(val)
                    if abs(val) > 1e-8:
                        vals = colored(vals, "red")
                    ss += " {eqn:2} : {vals} : {eqs}\n".format(
                        eqn=str(i + 1), vals=vals, eqs=eq
                    )
            ss += "\n"
        s += ss

        return s

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):

        from dolang.latex import eq2tex

        # general informations
        infos = self.infos
        table_infos = """
        <table>
         <td><b>Model</b></td>
        <tr>
        <td>name</td>
        <td>{name}</td>
        </tr>
        <tr>
        <td>type</td>
        <td>{type}</td>
        </tr>
        <tr>
        <td>filename</td>
        <td>{filename}</td>
        </tr>
        </table>""".format(
            name=infos["name"],
            type=infos["type"],
            filename=infos["filename"].replace("<", "&lt").replace(">", "&gt"),
        )

        # Equations and residuals
        resids = self.residuals()
        equations = self.equations.copy()
        # Create definitions equations and append to equations dictionary
        definitions = self.definitions
        tmp = []
        for deftype in definitions:
            tmp.append(deftype + " = " + definitions[deftype])

        definitions = {"definitions": tmp}
        equations.update(definitions)

        variables = sum([e for k, e in self.symbols.items() if k != "parameters"], [])
        table = '<tr><td><b>Type</b></td><td style="width:80%"><b>Equation</b></td><td><b>Residual</b></td></tr>\n'

        for eq_type in equations:

            eq_lines = []
            for i in range(len(equations[eq_type])):
                eq = equations[eq_type][i]
                # if eq_type in ('expectation','direct_response'):
                #     vals = ''
                if eq_type not in ("arbitrage", "transition", "arbitrage_exp"):
                    vals = ""
                else:
                    val = resids[eq_type][i]
                    if abs(val) > 1e-8:
                        vals = '<span style="color: red;">{:.4f}</span>'.format(val)
                    else:
                        vals = "{:.3f}".format(val)
                if "⟂" in eq:
                    # keep only lhs for now
                    eq, comp = str.split(eq, "⟂")
                if "|" in eq:
                    # keep only lhs for now
                    eq, comp = str.split(eq, "|")
                lat = eq2tex(variables, eq)
                lat = "${}$".format(lat)
                line = [lat, vals]
                h = eq_type if i == 0 else ""
                fmt_line = "<tr><td>{}</td><td>{}</td><td>{}</td></tr>".format(h, *line)
                #         print(fmt_line)
                eq_lines.append(fmt_line)
            table += str.join("\n", eq_lines)
        table = "<table>{}</table>".format(table)

        return table_infos + table

    @property
    def x_bounds(self):

        if "controls_ub" in self.functions:
            fun_lb = self.functions["controls_lb"]
            fun_ub = self.functions["controls_ub"]
            return [fun_lb, fun_ub]
        elif "arbitrage_ub" in self.functions:
            fun_lb = self.functions["arbitrage_lb"]
            fun_ub = self.functions["arbitrage_ub"]
            return [fun_lb, fun_ub]
        else:
            return None

    def residuals(self, calib=None):

        from dolo.algos.steady_state import residuals

        return residuals(self, calib)

    def eval_formula(self, expr, dataframe=None, calib=None):

        from dolo.compiler.eval_formula import eval_formula

        if calib is None:
            calib = self.calibration
        return eval_formula(expr, dataframe=dataframe, context=calib)



import ast
from ast import BinOp, Compare, Sub
import sympy
import numpy

from dolo.compiler.function_compiler_sympy import (
    ast_to_sympy,
    compile_higher_order_function,
)

from dolang.symbolic import stringify, stringify_symbol


def timeshift(expr, variables, date):
    from sympy import Symbol
    from dolang import stringify

    d = {
        Symbol(stringify_symbol((v, 0))): Symbol(stringify_symbol((v, date)))
        for v in variables
    }
    return expr.subs(d)


def parse_equation(eq_string, vars, substract_lhs=True, to_sympy=False):

    eq = eq_string.split("|")[0]  # ignore complentarity constraints

    if "==" not in eq:
        eq = eq.replace("=", "==")

    expr = ast.parse(eq).body[0].value
    expr_std = stringify(expr, variables=vars)

    if isinstance(expr_std, Compare):
        lhs = expr_std.left
        rhs = expr_std.comparators[0]
        if substract_lhs:
            expr_std = BinOp(left=rhs, right=lhs, op=Sub())
        else:
            if to_sympy:
                return [ast_to_sympy(lhs), ast_to_sympy(rhs)]
            return [lhs, rhs]

    if to_sympy:
        return ast_to_sympy(expr_std)
    else:
        return expr_std


def model_to_fg(model, order=2):
    # compile f, g function at higher order

    all_variables = sum(
        [model.symbols[e] for e in model.symbols if e != "parameters"], []
    )
    all_dvariables = (
        [(d, 0) for d in all_variables]
        + [(d, 1) for d in all_variables]
        + [(d, -1) for d in all_variables]
    )
    psyms = [(e, 0) for e in model.symbols["parameters"]]

    definitions = model.definitions

    d = dict()

    for k in definitions:
        v = parse_equation(definitions[k], all_variables, to_sympy=True)
        kk = stringify_symbol((k, 0))
        kk_m1 = stringify_symbol((k, -1))
        kk_1 = stringify_symbol((k, 1))
        d[sympy.Symbol(kk)] = v
        d[sympy.Symbol(kk_m1)] = timeshift(v, all_variables, -1)
        d[sympy.Symbol(kk_1)] = timeshift(v, all_variables, 1)

    f_eqs = model.equations["arbitrage"]
    f_eqs = [parse_equation(eq, all_variables, to_sympy=True) for eq in f_eqs]
    f_eqs = [eq.subs(d) for eq in f_eqs]

    g_eqs = model.equations["transition"]
    g_eqs = [
        parse_equation(eq, all_variables, to_sympy=True, substract_lhs=False)
        for eq in g_eqs
    ]
    # solve_recursively
    dd = dict()
    for eq in g_eqs:
        dd[eq[0]] = eq[1].subs(dd).subs(d)
    g_eqs = dd.values()

    f_syms = (
        [(e, 0) for e in model.symbols["states"]]
        + [(e, 0) for e in model.symbols["controls"]]
        + [(e, 1) for e in model.symbols["states"]]
        + [(e, 1) for e in model.symbols["controls"]]
    )

    g_syms = (
        [(e, -1) for e in model.symbols["states"]]
        + [(e, -1) for e in model.symbols["controls"]]
        + [(e, 0) for e in model.symbols["exogenous"]]
    )

    params = model.symbols["parameters"]

    f = compile_higher_order_function(
        f_eqs,
        f_syms,
        params,
        order=order,
        funname="f",
        return_code=False,
        compile=False,
    )

    g = compile_higher_order_function(
        g_eqs,
        g_syms,
        params,
        order=order,
        funname="g",
        return_code=False,
        compile=False,
    )
    # cache result
    model.__higher_order_functions__ = dict(f=f, g=g)
    model.__highest_order__ = order

    return [f, g]


def get_model_derivatives(model, order, calibration=None):

    if calibration is None:
        calibration = model.calibration

    [f_fun, g_fun] = model_to_fg(model, order=order)

    parms = model.calibration["parameters"]
    states_ss = calibration["states"]
    controls_ss = calibration["controls"]
    shocks_ss = calibration["exogenous"]

    f_args_ss = numpy.concatenate([states_ss, controls_ss, states_ss, controls_ss])
    g_args_ss = numpy.concatenate([states_ss, controls_ss, shocks_ss])

    f = f_fun(f_args_ss, parms, order=order)
    g = g_fun(g_args_ss, parms, order=order)

    return [f, g]



from __future__ import division

import numpy
from dolo.numeric.misc import cartesian

# Credits : both routines below are ported from the Compecon Toolbox
# by Paul L Fackler and Mario J. Miranda.
# It is downloadable at http://www4.ncsu.edu/~pfackler/compecon/toolbox.html
def hermgauss(n):

    from numpy import pi, fix, zeros, sqrt

    maxit = 100
    pim4 = 1 / pi ** 0.25
    m = int(fix((n + 1) / 2))

    x = zeros(n)
    w = zeros(n)
    # reasonable starting values
    for i in range(m):
        if i == 0:
            z = sqrt(2 * n + 1) - 1.85575 * ((2 * n + 1) ** (-1 / 6))
        elif i == 1:
            z = z - 1.14 * (n ** 0.426) / z
        elif i == 2:
            z = 1.86 * z + 0.86 * x[0]
        elif i == 3:
            z = 1.91 * z + 0.91 * x[1]
        else:
            z = 2 * z + x[i - 2]
        # root finding iterations
        its = 0
        while its < maxit:
            its += 1
            p1 = pim4
            p2 = 0
            for j in range(n):
                p3 = p2
                p2 = p1
                p1 = z * sqrt(2 / (j + 1)) * p2 - sqrt(j / (j + 1)) * p3
            pp = sqrt(2 * n) * p2
            z1 = z
            z = z1 - p1 / pp
            if abs(z - z1) < 1e-14:
                break
        if its >= maxit:
            raise Exception("Failure to converge")
        x[n - i - 1] = z
        x[i] = -z
        w[i] = 2 / pp ** 2
        w[n - i - 1] = w[i]

    return [x, w]


def gauss_hermite_nodes(orders, sigma, mu=None):
    """
    Computes the weights and nodes for Gauss Hermite quadrature.

    Parameters
    ----------
    orders : int, list, array
        The order of integration used in the quadrature routine
    sigma : array-like
        If one dimensional, the variance of the normal distribution being
        approximated. If multidimensional, the variance-covariance matrix of
        the multivariate normal process being approximated.

    Returns
    -------
    x : array
        Quadrature nodes
    w : array
        Quadrature weights
    """
    if isinstance(orders, int):
        orders = [orders]

    import numpy

    if mu is None:
        mu = numpy.array([0] * sigma.shape[0])

    herms = [hermgauss(i) for i in orders]

    points = [h[0] * numpy.sqrt(2) for h in herms]
    weights = [h[1] / numpy.sqrt(numpy.pi) for h in herms]

    if len(orders) == 1:
        # Note: if sigma is 2D, x will always be 2D, even if sigma is only 1x1.
        # print(points.shape)
        x = numpy.array(points[0]) * numpy.sqrt(float(sigma))
        if sigma.ndim == 2:
            x = x[:, None]
        w = weights[0]
        return [x, w]

    else:
        x = cartesian(points).T

        from functools import reduce

        w = reduce(numpy.kron, weights)

        zero_columns = numpy.where(sigma.sum(axis=0) == 0)[0]
        for i in zero_columns:
            sigma[i, i] = 1.0

        C = numpy.linalg.cholesky(sigma)

        x = numpy.dot(C, x) + mu[:, numpy.newaxis]

        x = numpy.ascontiguousarray(x.T)

        for i in zero_columns:
            x[:, i] = 0

        return [x, w]


# from numpy.polynomial.hermite import hermgauss


if __name__ == "__main__":

    orders = [8, 8]
    mu = numpy.array([0.05, 0.01])
    sigma = numpy.array(
        [
            [0.1, 0.015],
            [0.015, 0.1],
        ]
    )

    from numpy.polynomial.hermite import hermgauss as hermgauss_numpy

    [xg, wg] = hermgauss_numpy(10)
    [x, w] = hermgauss(10)
    print(w - wg)
    print(x - xg)



import unittest
import numpy


# def test_web_import():
#
#     from dolo import yaml_import
#
#     model = yaml_import("https://raw.githubusercontent.com/EconForge/dolo/master/examples/models/rbc.yaml")
#     assert(len(model.symbols['states'])==2)


def model_evaluation(compiler="numpy", data_layout="columns"):

    from dolo import yaml_import

    model = yaml_import("examples/models/rbc_iid.yaml")

    s0 = model.calibration["states"]
    x0 = model.calibration["controls"]
    e0 = model.calibration["exogenous"]
    p = model.calibration["parameters"]

    assert s0.ndim == 1
    assert x0.ndim == 1
    assert p.ndim == 1

    N = 10
    f = model.functions["arbitrage"]

    if data_layout == "rows":
        ss = numpy.ascontiguousarray(numpy.tile(s0, (N, 1)).T)
        xx = numpy.ascontiguousarray(numpy.tile(x0, (N, 1)).T)
        ee = e0[:, None].repeat(N, axis=1)
    else:
        ss = numpy.tile(s0, (N, 1))
        xx = numpy.tile(x0, (N, 1))
        ee = e0[None, :].repeat(N, axis=0)

    ss = s0[None, :].repeat(N, axis=0)
    xx = x0[None, :].repeat(N, axis=0)
    ee = e0[None, :].repeat(N, axis=0)

    vec_res = f(ee, ss, xx, ee, ss, xx, p)

    res = f(e0, s0, x0, e0, s0, x0, p)

    assert res.ndim == 1

    d = 0
    for i in range(N):
        d += abs(vec_res[i, :] - res).max()
    assert d == 0


if __name__ == "__main__":
    model_evaluation()



import numpy as np

# doesn't seem to support keyword arguments yet
# from multipledispatch import dispatch
# multimethod = dispatch()
from dolo.misc.multimethod import multimethod
from numba import generated_jit
import xarray

from dolo.compiler.model import Model
from dolo.numeric.grids import UniformCartesianGrid
from dolo.numeric.decision_rule import DecisionRule
from dolo.numeric.processes import MarkovChain, IIDProcess, DiscretizedIIDProcess
from dolo.numeric.grids import CartesianGrid, UnstructuredGrid, EmptyGrid


@generated_jit(nopython=True)
def trembling_hand(A: "N*n1*...*nd", x: "N*d", w: "float"):

    # we could generate that automatically for orders >=3
    if A.ndim == 2:

        def fun(A: "N* *...", x: "N*d", w: "float"):

            N, n0 = A.shape
            δ0 = 1 / (n0 - 1)

            for n in range(N):

                x0 = x[n, 0]
                x0 = min(max(x0, 0), 1)
                q0 = np.floor_divide(x0, δ0)
                q0 = max(0, q0)
                q0 = min(q0, n0 - 2)

                λ0 = x0 / δ0 - q0  # ∈[0,1[ by construction

                q0_ = int(q0)

                A[n, q0_] += (1 - λ0) * w
                A[n, q0_ + 1] += λ0 * w

        return fun

    elif A.ndim == 3:

        def fun(A: "N* *...", x: "N*d", w: "float"):

            N, n0, n1 = A.shape
            δ0 = 1 / (n0 - 1)
            δ1 = 1 / (n1 - 1)

            for n in range(N):

                x0 = x[n, 0]
                x0 = min(max(x0, 0), 1)
                q0 = np.floor_divide(x0, δ0)
                q0 = max(0, q0)
                q0 = min(q0, n0 - 2)

                x1 = x[n, 1]
                x1 = min(max(x1, 0), 1)
                q1 = np.floor_divide(x1, δ1)
                q1 = max(0, q1)
                q1 = min(q1, n1 - 2)

                λ0 = (x0 - q0 * δ0) / δ0  # ∈[0,1[ by construction
                q0_ = int(q0)

                λ1 = (x1 - q1 * δ1) / δ1  # ∈[0,1[ by construction
                q1_ = int(q1)

                A[n, q0_, q1_] += (1 - λ0) * (1 - λ1) * w
                A[n, q0_, q1_ + 1] += (1 - λ0) * (λ1) * w
                A[n, q0_ + 1, q1_] += (λ0) * (1 - λ1) * w
                A[n, q0_ + 1, q1_ + 1] += (λ0) * (λ1) * w

        return fun

    else:
        raise Exception("Unsupported.")


# TODO: add default options for endo_grid, exo_grid, dp
@multimethod
def ergodic_distribution(model: Model, dr: DecisionRule):
    return ergodic_distribution(model, dr, dr.exo_grid, dr.endo_grid, dr.dprocess)


@multimethod
def ergodic_distribution(
    model: Model,
    dr: DecisionRule,
    exo_grid: UnstructuredGrid,
    endo_grid: UniformCartesianGrid,
    dp: MarkovChain,
    compute_μ=True,
):

    N_m = exo_grid.n_nodes
    N_s = endo_grid.n_nodes
    dims_s = tuple(endo_grid.n)  # we have N_s = prod(dims_s)
    s = endo_grid.nodes

    N = N_m * N_s
    Π = np.zeros((N_m, N_m, N_s) + dims_s)

    g = model.functions["transition"]
    p = model.calibration["parameters"]

    a = endo_grid.min
    b = endo_grid.max

    for i_m in range(N_m):
        m = exo_grid.node(i_m)
        x: np.array = dr(i_m, s)
        for i_M in range(dp.n_inodes(i_m)):
            w = dp.iweight(i_m, i_M)
            M = dp.inode(i_m, i_M)
            S = g(m, s, x, M, p)

            # renormalize
            S = (S - a[None, :]) / (b[None, :] - a[None, :])

            # allocate weights
            dest = Π[i_m, i_M, ...]
            trembling_hand(dest, S, w)

    # these seem to be a bit expensive...
    Π = Π.reshape((N_m, N_m, N_s, N_s))
    Π = Π.swapaxes(1, 2)
    Π = Π.reshape((N, N))

    if not compute_μ:
        return Π.reshape((N_m, N_s, N_m, N_s))
    else:
        B = np.zeros(N)
        B[-1] = 1.0
        A = Π.T - np.eye(Π.shape[0])
        A[-1, :] = 1.0
        μ = np.linalg.solve(A, B)

        μ = μ.reshape((N_m,) + dims_s)
        labels = [
            np.linspace(endo_grid.min[i], endo_grid.max[i], endo_grid.n[i])
            for i in range(len(endo_grid.max))
        ]
        μ = xarray.DataArray(
            μ,
            [("i_m", np.arange(N_m))]
            + list(
                {s: labels[i] for i, s in enumerate(model.symbols["states"])}.items()
            ),
        )
        return Π.reshape((N_m, N_s, N_m, N_s)), μ


@multimethod
def ergodic_distribution(
    model: Model,
    dr: DecisionRule,
    exo_grid: EmptyGrid,
    endo_grid: CartesianGrid,
    dp: DiscretizedIIDProcess,
    compute_μ=True,
):

    N_s = endo_grid.n_nodes
    dims_s = tuple(endo_grid.n)  # we have N_s = prod(dims_s)
    s = endo_grid.nodes

    m = model.calibration["exogenous"]
    Π = np.zeros((N_s,) + dims_s)

    g = model.functions["transition"]
    p = model.calibration["parameters"]

    a = endo_grid.min
    b = endo_grid.max

    x: np.array = dr(s)

    for i_M in range(dp.n_inodes(0)):

        w = dp.iweight(0, i_M)
        M = dp.inode(0, i_M)
        S = g(m, s, x, M, p)
        # renormalize
        S = (S - a[None, :]) / (b[None, :] - a[None, :])
        # allocate weights
        trembling_hand(Π, S, w)

    Π = Π.reshape((N_s, N_s))

    if not compute_μ:
        return Π

    else:
        B = np.zeros(N_s)
        B[-1] = 1.0
        A = Π.T - np.eye(Π.shape[0])
        A[-1, :] = 1.0
        μ = np.linalg.solve(A, B)
        μ = μ.reshape(dims_s)

        labels = [
            np.linspace(endo_grid.min[i], endo_grid.max[i], endo_grid.n[i])
            for i in range(len(endo_grid.max))
        ]
        μ = xarray.DataArray(
            μ,
            list({s: labels[i] for i, s in enumerate(model.symbols["states"])}.items()),
        )

        return Π, μ



# this is for compatibility purposes only

from dolo.algos.time_iteration import time_iteration
from dolo.algos.perfect_foresight import deterministic_solve
from dolo.algos.simulations import simulate, response, tabulate, plot_decision_rule
from dolo.algos.value_iteration import evaluate_policy, value_iteration
from dolo.algos.improved_time_iteration import improved_time_iteration
from dolo.algos.steady_state import residuals
from dolo.algos.perturbation import perturb
from dolo.algos.ergodic import ergodic_distribution

approximate_controls = perturb



import numpy as np


def cartesian(arrays, out=None):
    """
    Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n // arrays[0].size
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j * m : (j + 1) * m, 1:] = out[0:m, 1:]
    return out


def mlinspace(a, b, orders, out=None):

    import numpy

    sl = [numpy.linspace(a[i], b[i], orders[i]) for i in range(len(a))]

    if out is None:
        out = cartesian(sl)
    else:
        cartesian(sl, out)

    return out


def MyJacobian(fun, eps=1e-6):
    def rfun(x):
        n = len(x)
        x0 = x.copy()
        y0 = fun(x)
        D = np.zeros((len(y0), len(x0)))
        for i in range(n):
            delta = np.zeros(len(x))
            delta[i] = eps
            y1 = fun(x + delta)
            y2 = fun(x - delta)
            D[:, i] = (y1 - y2) / eps / 2
        return D

    return rfun



## Useful Links

## Common probability distributions:
## https://blog.cloudera.com/blog/2015/12/common-probability-distributions-the-data-scientists-crib-sheet/
##
## Distributions.jl:(args, kwargs))
## https://juliastats.github.io/Distributions.jl/stable/
##
## Scipy.stats:
## https://docs.scipy.org/doc/scipy-0.17.1/reference/stats.html
##
## Quantecon/rvlib list:
## https://github.com/QuantEcon/rvlib/tree/multivariate
##
## Quantecon/rvlib univarite:
## https://github.com/QuantEcon/rvlib/blob/multivariate/rvlib/univariate.py
##
## Hark/utilities.py: (all univariate?)
## https://github.com/econ-ark/HARK/blob/d99393973554b1cf830c6285e6da59d98ff242ff/HARK/utilities.py
##
## Dolo processes.py:
## https://github.com/EconForge/dolo/blob/master/dolo/numeric/processes.py
##
## Dolo processes.jl:
## https://github.com/EconForge/Dolo.jl/blob/master/src/numeric/processes.jl


## This code

# Here we have the list of classes implemented by Rvlib
# They only have Mv normal in multivariate (add at least log-normal)

# Do we sepearte mv and univariate?
# If not create a dict with small and capital letters denoting uni and mv cases
# Then accept both and convert automatically to do operations for mv case ?
# For which cases we have mv? mostly Normal/UNormal

# Parameter names are (so far) used as they appear in Distributions.jl
# This seems like the richest source for distributions with very clear documentation


import numpy as np  # type: ignore
from scipy.stats import norm, uniform, lognorm, beta  # type: ignore
from matplotlib import pyplot as plt  # type: ignore
import numpy as np  # type: ignore
from dataclasses import dataclass  # type: ignore

from typing import List, TypeVar, Generic, Union, Any, Callable  # type: ignore
from typing import Iterator, Tuple  # type: ignore

from dolang.language import greek_tolerance, language_element  # type: ignore
from dolo.numeric.processes import IIDProcess, DiscretizedIIDProcess, MarkovChain  # type: ignore

Vector = List[float]
Matrix = List[Vector]
T = TypeVar("T")


class Distribution(IIDProcess):
    """
    A multivariate distribution.

    Attributes:
        d(int): number of dimensions.
        names(list[str], optional): variable names
    """

    d: int  # number of dimensions
    names: Union[None, Tuple[str, ...]]  # names of variables (optional)

    def draw(self, N: int) -> Matrix:
        "Compute `N` random draws. Returns an `N` times `d` matrix."

        raise Exception(
            f"Not Implemented (yet). Should be implemented by subclass {self.__class__}"
        )

    def integrate(self, f) -> float:
        "Computes the expectation $E_u f(u)$ for given function `f`"

        raise Exception(
            f"Not Implemented (yet). Should be implemented by subclass {self.__class__}"
        )


###
### Continuous Distributions
###


class ContinuousDistribution(Distribution):
    def discretize(self, **kwargs):  # ->DiscreteDistribution:

        raise Exception(
            f"Not Implemented (yet). Should be implemented by subclass {self.__class__}"
        )


###
### Discrete Distributions
###


class DiscreteDistribution(Distribution, DiscretizedIIDProcess):
    """
    A multivariate discrete distribution.

    Attributes:
        d(int): number of dimensions.
        names(list[str], optional): variable names
        n(int):  number of discretization points
        origin(distribution, optional): distribution that was discretized
    """

    n: int  # number of discretization points
    origin: Union[None, ContinuousDistribution]

    def point(self, i) -> Vector:
        "Returns i-th discretization point (a Vector)"
        raise Exception(
            f"Not Implemented (yet). Should be implemented by subclass {self.__class__}"
        )

    def weight(self, i) -> float:
        "Returns i-th discretization point (a float)"
        raise Exception(
            f"Not Implemented (yet). Should be implemented by subclass {self.__class__}"
        )

    def items(self) -> Iterator[Tuple[float, Vector]]:
        """Returns a generator yielding all points and weights.

        Example: sum( [ w*f(x) for (w,x) in discrete_dist.items() ] )
        """
        return ((self.weight(i), self.point(i)) for i in range(self.n))

    def integrate(self, fun: Callable[[Vector], T]) -> T:
        # alread documented by the ancestor
        return sum(w * fun(x) for (w, x) in self.items())


class EquiprobableDistribution(DiscreteDistribution):

    points: Vector

    def __init__(self, points: Vector = None, origin: Union[Distribution, None] = None):

        n, d = points.shape
        self.d = d
        self.n = n
        self.points = points
        self.origin = origin

    @property
    def weights(self) -> Vector:
        # so that it can behave like a FiniteDistribution (notably for graphs)
        w = np.ones(self.n)
        w /= self.n
        return w

    def point(self, i) -> float:

        return self.points[i, :]

    def weight(self, i) -> float:

        return 1 / self.n

    def draw(self, N: int) -> Matrix:
        import numpy.random

        inds = numpy.random.randint(low=0, high=self.n, size=N)
        return self.points[inds, :]

    def discretize(self, to="iid"):
        if to == "iid":
            return self
        elif to == "mc":
            return FiniteDistribution(self.points, self.weights).discretize(to="mc")
        else:
            raise Exception("Not implemented.")

    def __repr__(self):
        return f"EquiprobableDistribution(points={self.points.__repr__()}, origin={str(self.origin)})"

    def __str__(self):
        return f"EquiprobableDistribution(points={self.points}, origin={self.origin})"


# Special kind of Discrete distributions characterized
# by a list of points and a list of weights.
class FiniteDistribution(DiscreteDistribution):

    points: Vector
    weights: Vector

    def __init__(
        self,
        points: Vector = None,
        weights: Vector = None,
        origin: Union[Distribution, None] = None,
    ):

        n, d = points.shape
        self.d = d
        self.n = n
        assert len(weights) == n
        self.points = points
        self.weights = weights
        self.origin = origin

    def draw(self, N: int) -> Matrix:

        import numpy.random

        choices = numpy.random.choice(range(self.n), size=N, p=self.weights)
        return self.points[choices, :]

    def point(self, i) -> float:

        return self.points[i, :]

    def weight(self, i) -> float:

        return self.weights[i]

    def discretize(self, to="iid"):
        if to == "iid":
            return self
        elif to == "mc":
            from .processes import MarkovChain

            nodes = self.points
            N = len(nodes)
            transitions = np.array(
                [
                    self.weights,
                ]
                * N
            )
            return MarkovChain(transitions, nodes)
        else:
            raise Exception("Not implemented.")

    def __repr__(self):
        return f"FiniteDistribution(points={self.points.__repr__()}, weights={self.weights.__repr__()}, origin={str(self.origin)})"

    def __str__(self):
        return f"FiniteDistribution(points={self.points}, weights={self.weights}, origin={self.origin})"


def product_iid(iids: List[FiniteDistribution]) -> FiniteDistribution:

    from dolo.numeric.misc import cartesian

    nn = [len(f.weights) for f in iids]

    cart = cartesian([range(e) for e in nn])

    nodes = np.concatenate(
        [f.points[cart[:, i], :] for i, f in enumerate(iids)], axis=1
    )
    weights = iids[0].weights
    for f in iids[1:]:
        weights = np.kron(weights, f.weights)

    return FiniteDistribution(nodes, weights)


###
### Discrete Distributions
###


@language_element
##@dataclass
class Bernouilli(DiscreteDistribution):

    π: float = 0.5

    signature = {"π": "float"}  # this is redundant for now

    @greek_tolerance
    def __init__(self, π: float = None):
        self.π = float(π)

    def discretize(self, to="iid"):

        if to == "iid":
            x = np.array([[0], [1]])
            w = np.array([1 - self.π, self.π])
            return FiniteDistribution(x, w)
        elif to == "mc":
            fin_distr = self.discretize(to="iid")
            return fin_distr.discretize(to="mc")
        else:
            raise Exception("Not implemented.")

    def draw(self, N: int) -> Matrix:
        a = np.array([0, 1])
        c = np.random.choice(a, size=N)
        return c.reshape((N, 1))


@language_element
##@dataclass
class Binomial(DiscreteDistribution):

    π: float = 0.5
    n: int

    # TODO


###
### 1d Continuous Distributions
###


class UnivariateContinuousDistribution(ContinuousDistribution):
    """
    A univariate distribution.

    Attributes:
        d(int): number of dimensions.
        names(list[str], optional): variable names
    """

    d = 1

    def ppf(self, quantiles: Vector) -> Vector:
        "Percentage Point Function (inverse CDF)"
        raise Exception(
            f"Not Implemented (yet). Should be implemented by subclass {self.__class__}"
        )

    def cdf(self, quantiles: Vector) -> Vector:
        "Cumulative Distribution"
        raise Exception(
            f"Not Implemented (yet). Should be implemented by subclass {self.__class__}"
        )

    def discretize(self, to="iid", N=5, method="equiprobable", mass_point="median"):
        """
        Returns a discretized version of this process.

        Parameters
        ----------
        N : int
            Number of point masses in the discretized distribution.

        method : str
            'equiprobable' or 'gauss-hermite'

        mass_point : str
            'median', 'left', 'middle', or 'right'


        Returns:
        ------------
        process : DiscreteDistribution
            A discrete distribution.
        """
        if to == "iid":
            if method == "gauss-hermite":
                return self.__discretize_gh__(N=N)
            elif method == "equiprobable":
                return self.__discretize_ep__(N=N, mass_point=mass_point)
            else:
                raise Exception("Unknown discretization method.")
        elif to == "mc":
            discr_iid = self.discretize(to="iid")
            return discr_iid.discretize(to="mc")
        else:
            raise Exception("Not implemented (yet).")

    def __discretize_ep__(self, N=5, mass_point="median"):  # Equiprobable
        if mass_point == "median":
            p = np.linspace(0.5 / N, 1 - 0.5 / N, N)
            q = self.ppf(p)
        elif mass_point == "left":
            p = np.linspace(0, 1 - 1 / N, N)
            q = self.ppf(p)
        elif mass_point == "middle":
            p = np.linspace(0.0, 1, N + 1)
            q = self.ppf(p)
            q = 0.5 * (q[1:] + q[:-1])
        elif mass_point == "right":
            p = np.linspace(1 / N, 1, N)
            q = self.ppf(p)
        else:
            raise Exception("Not implemented")

        w = (1 / (N)) * np.ones(N)

        return EquiprobableDistribution(q[:, None], origin=self)


@language_element
##@dataclass
class UNormal(UnivariateContinuousDistribution):

    μ: float = 0.0
    σ: float = 1.0

    signature = {"μ": "Optional[float]", "σ": "float"}  # this is redundant for now

    @greek_tolerance
    def __init__(self, σ: float = None, μ: float = None):
        self.σ = float(σ)
        self.μ = 0.0 if μ is None else float(μ)

    def ppf(self, quantiles):
        x = norm.ppf(quantiles, loc=self.μ, scale=self.σ)
        return x

    def cdf(self, x):
        p = norm.cdf(x, loc=self.μ, scale=self.σ)
        return p

    def draw(self, N):

        from numpy.random import multivariate_normal

        Sigma = np.array([[self.σ ** 2]])
        mu = np.array([self.μ])
        sim = multivariate_normal(mu, Sigma, size=N)
        return sim.reshape((N, 1))

    def integrate(self, fun) -> float:

        # I don't think anybody should use that. It's just an example
        σ = self.σ
        μ = self.μ

        f = (
            lambda x: fun(x)
            / σ
            / np.sqrt(2 * np.pi)
            * np.exp(-1 / 2 * ((x - μ) / σ) ** 2)
        )
        import scipy.integrate

        return scipy.integrate.quad(f, -np.Inf, np.Inf)[0]

    def __discretize_gh__(self, N=5):  # Gauss-Hermite

        # Maybe we can avoid that one by inheriting from mvNormal
        from dolo.numeric.discretization.quadrature import gauss_hermite_nodes

        [x, w] = gauss_hermite_nodes(N, np.array([[self.σ ** 2]]), mu=self.μ)
        x += np.array([self.μ])[:, None]
        return FiniteDistribution(x, w, origin=self)


@language_element
##@dataclass
class Uniform(UnivariateContinuousDistribution):

    # uniform distribution over an interval [a,b]
    a: float = 0.0
    b: float = 1.0

    signature = {"a": "float", "b": "float"}  # this is redundant for now

    def __init__(self, a: float = 0.0, b: float = 1.0):
        self.a = float(a)
        self.b = float(b)

    def ppf(self, quantiles: Vector) -> Vector:
        x = uniform.ppf(quantiles, loc=self.a, scale=(self.b - self.a))
        return x

    def cdf(self, x: Vector) -> Vector:
        p = uniform.cdf(x, loc=self.a, scale=(self.b - self.a))
        return p

    def draw(self, N) -> Matrix:
        from numpy.random import uniform

        sim = uniform(self.a, self.b, N)
        return sim.reshape((N, 1))


@language_element
# @dataclass
class LogNormal(UnivariateContinuousDistribution):

    # parametrization a lognormal random variable Y is in terms of
    # the mean, μ, and standard deviation, σ, of the unique normally distributed random variable X
    # such that exp(X) = Y.

    μ: float = 0.0
    σ: float = 1.0

    signature = {"μ": "Optional[float]", "σ": "float"}  # this is redundant for now

    @greek_tolerance
    def __init__(self, σ: float = 0.0, μ: float = 1.0):
        self.σ = float(σ)
        self.μ = float(μ)

    # From scipy: defined as lognorm.pdf(x, s, loc, scale)
    # See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.lognorm.html#scipy.stats.lognorm

    # A common parametrization for a lognormal random variable Y is in terms of
    # the mean, mu, and standard deviation, sigma, of the unique normally distributed random variable X
    # such that exp(X) = Y.
    # This parametrization corresponds to setting s = sigma and scale = exp(mu).

    def ppf(self, quantiles):
        x = lognorm.ppf(quantiles, s=self.σ, loc=0, scale=np.exp(self.μ))
        return x

        # method: See hark
        # Makes an equiprobable distribution by
        # default, but user can optionally request augmented tails with exponentially
        # sized point masses.  This can improve solution accuracy in some models.
        # https://github.com/econ-ark/HARK/blob/d99393973554b1cf830c6285e6da59d98ff242ff/HARK/utilities.py

        # Returns
        # inodes
        # Discrete points for discrete probability mass function.
        # weights
        # Probability associated with each point in grid (nodes)


class Beta(UnivariateContinuousDistribution):

    α: float
    β: float

    signature = {"α": "float", "β": "float"}

    def __init__(self, α: float = None, β: float = None):
        self.α = float(α)
        self.β = float(β)

    def ppf(self, quantiles):
        x = beta.ppf(quantiles, self.α, self.β)
        return x

        # method: In hark utilities
        # reference https://github.com/econ-ark/HARK/blob/d99393973554b1cf830c6285e6da59d98ff242ff/HARK/utilities.py

        # Returns
        # inodes ("X" in hark)
        # Discrete points for discrete probability mass function.
        # weights ("pmf" in Hark : Discrete points for discrete probability mass function.)
        # Probability associated with each point in grid (nodes)


#%


###
### nd continous distribution
###


@language_element
class Normal(ContinuousDistribution):

    Μ: Vector  # this is capital case μ, not M... 😭
    Σ: Matrix

    signature = {"Σ": "Matrix", "Μ": "Optional[Vector]"}

    @greek_tolerance
    def __init__(self, Σ=None, Μ=None):

        Sigma = Σ
        mu = Μ

        self.Σ = np.atleast_2d(np.array(Sigma, dtype=float))
        self.d = len(self.Σ)
        if mu is None:
            self.Μ = np.array([0.0] * self.d)
        else:
            self.Μ = np.array(mu, dtype=float)
        assert self.Σ.shape[0] == self.d
        assert self.Σ.shape[0] == self.d

        # this class wraps functionality from scipy
        import scipy.stats

        self._dist_ = scipy.stats.multivariate_normal(mean=self.Μ, cov=self.Σ)

    def draw(self, N: int) -> Matrix:

        res = self._dist_.rvs(size=N)
        if res.ndim == 1:
            # scipy returns a 1d object for 1d distribution
            res = res[:, None]

        return res

    def discretize(self, to="iid", N=None) -> FiniteDistribution:

        if to == "iid":
            if N is None:
                N = 5
            if isinstance(N, int):
                N = [N] * self.d
            from dolo.numeric.discretization.quadrature import gauss_hermite_nodes  # type: ignore

            [x, w] = gauss_hermite_nodes(N, self.Σ, mu=self.Μ)
            x = np.row_stack([(e + self.Μ) for e in x])
            return FiniteDistribution(x, w, origin=self)

        elif to == "mc":
            discr_iid = self.discretize(to="iid")
            return discr_iid.discretize(to="mc")

        else:
            raise Exception("Not implemented.")

    def __repr__(self):
        return f"Normal(Μ={self.Μ.__repr__()},Σ={self.Σ.__repr__()})"

    def __str__(self):
        return f"Normal(Μ={self.Μ},Σ={self.Σ})"


MvNormal = Normal


class ProductDistribution(ContinuousDistribution):

    distributions: List[Distribution]

    # def __new__(self, distributions: List[Distributions]):
    #     # if all distributions are normal we can interrupt the object
    #     # construction and return a multivariate normal object instead
    #     # of a product object

    def __init__(self, distributions: List[Distribution]):

        self.distributions = distributions
        self.d = sum([dis.d for dis in distributions])
        self.names = sum([dis.names for dis in self.distributions], tuple())

    def discretize(self, to="iid"):

        # TODO: pass some options
        fids = [dis.discretize(to=to) for dis in self.distributions]
        return product_iid(fids)

    def draw(self, N: int) -> Matrix:

        return np.concatenate([dis.draw(N) for dis in self.distributions], axis=1)


def product_iid(iids: List[FiniteDistribution]) -> FiniteDistribution:

    from dolo.numeric.misc import cartesian

    nn = [len(f.weights) for f in iids]

    cart = cartesian([range(e) for e in nn])

    nodes = np.concatenate(
        [f.points[cart[:, i], :] for i, f in enumerate(iids)], axis=1
    )
    weights = iids[0].weights
    for f in iids[1:]:
        weights = np.kron(weights, f.weights)

    return FiniteDistribution(nodes, weights)


###
### Truncation and Mixtures
###

C = TypeVar("C", bound=ContinuousDistribution)
C1 = TypeVar("C1", bound=ContinuousDistribution)
C2 = TypeVar("C2", bound=ContinuousDistribution)


class Truncation(UnivariateContinuousDistribution, Generic[C]):

    dist: C

    def __init__(self, dist: C, lb=-np.inf, ub=np.inf):

        self.dist = dist
        if lb == -np.inf:
            self.__min_q__ = 0.0
        else:
            self.__min_q__ = self.dist.cdf([lb])[0]
        if ub == np.inf:
            self.__max_q__ = 1.0
        else:
            self.__max_q__ = self.dist.cdf([ub])[0]

    def draw(self, N: int):

        # TODO: replace this stupid algo
        raise Exception("Not Implemented")

    def ppf(self, quantiles: Vector) -> Vector:

        q_lb = self.__min_q__
        q_ub = self.__max_q__

        q = q_lb + (q_ub - q_lb) * quantiles
        return self.dist.ppf(q)


@language_element
class Mixture(ContinuousDistribution):

    index: DiscreteDistribution  # values must be [0,1,..n]
    distributions: Tuple[UnivariateContinuousDistribution, ...]  # length musth be [n]

    signature = {"index": "DiscreteDistribution", "distributions": "List[Distribution]"}

    def __init__(self, index=None, distributions=None):
        # index is a distribution which takes discrete values
        # distributions is a map from each of these values to a distribution
        self.index = index
        self.distributions = distributions
        ds = [e.d for e in self.distributions.values()]
        assert len(set(ds)) == 1
        d0 = [*self.distributions.values()][0]
        self.d = d0.d
        # TODO: check all distributions have the same variable names
        self.names = d0.names

    def discretize(self, to="iid"):

        if to == "iid":

            inddist = self.index.discretize(to=to)
            nodes = []
            weights = []
            for i in range(inddist.n_inodes(0)):
                wind = inddist.iweight(0, i)
                xind = inddist.inode(0, i)
                dist = self.distributions[str(i)].discretize(to=to)
                for j in range(dist.n_inodes(0)):
                    w = dist.iweight(0, j)
                    x = dist.inode(0, j)
                    nodes.append(x)
                    weights.append(wind * w)
            nodes = np.concatenate([e[None, :] for e in nodes], axis=0)
            weights = np.array(weights)
            return FiniteDistribution(nodes, weights)

        elif to == "mc":
            from dolo.numeric.processes import DiscretizedIIDProcess

            return self.discretize(to="iid").discretize(to="mc")

        else:
            raise Exception("Not implemented.")

    def draw(self, N: int) -> Matrix:

        # naive and overkill algorithm
        inds = self.index.draw(N)  # should be (N x 1) array
        return sum(
            [(inds == k) * dist.draw(N) for (k, dist) in self.distributions.items()]
        )


# @language_element
# def Mixture(index=None, distributions=None):

#     for dist in distributions.values():
#         if not (isinstance(dist, IIDProcess)):
#             raise Exception("Only mixtures of iid processes are supported so far.")
#     return IIDMixture(index, distributions)
#     # not clear what we might do with non-iid

# Mixture.signature = {'index': 'intprocess', 'distributions': 'Dict[int,IIDProcesses]'}



from dolo.compiler.model import Model
from typing import Dict, List
import numpy as np

from dolo.compiler.misc import CalibrationDict


def residuals(model: Model, calib=None) -> Dict[str, List[float]]:

    if calib is None:
        calib = model.calibration

    res = dict()

    m = calib["exogenous"]
    s = calib["states"]
    x = calib["controls"]
    p = calib["parameters"]
    f = model.functions["arbitrage"]
    g = model.functions["transition"]

    res["transition"] = g(m, s, x, m, p) - s
    res["arbitrage"] = f(m, s, x, m, s, x, p)

    return res


def find_steady_state(model, m=None):

    n_s = len(model.calibration["states"])
    n_x = len(model.calibration["controls"])

    if m is None:
        m = model.calibration["exogenous"]
    p = model.calibration["parameters"]

    def fobj(v):
        s = v[:n_s]
        x = v[n_s:]
        d = dict(states=s, controls=x, exogenous=m, parameters=p)
        res = residuals(model, d)
        return np.concatenate([res["transition"], res["arbitrage"]])

    calib = model.calibration
    x0 = np.concatenate([calib["states"], calib["controls"]])
    import scipy.optimize

    sol = scipy.optimize.root(fobj, x0)
    res = sol.x

    d = dict(exogenous=m, states=res[:n_s], controls=res[n_s:], parameters=p)
    return CalibrationDict(model.symbols, d)



# from __future__ import print_function

# This module is supposed to be imported first
# it contains global variables used for configuration


# try to register printing methods if IPython is running

save_plots = False
real_type = "double"
debug = False

import warnings

from dolo.misc.termcolor import colored


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return "{}:{}:{}\n    {}\n".format(
        colored(category.__name__, "yellow"), filename, lineno, message
    )


warnings.formatwarning = warning_on_one_line
import numpy

numpy.seterr(all="ignore")


# create temporary directory for compiled functions
import tempfile, sys

temp_dir = tempfile.mkdtemp(prefix="dolo_")
sys.path.append(temp_dir)


from IPython.core.display import display
